PUBLIC_REPOSITORY_URL = "https://pkgs.safetycli.com/repository/public/pypi/simple/"
ORGANIZATION_REPOSITORY_URL = "https://pkgs.safetycli.com/repository/{}/pypi/simple/"
PROJECT_CONFIG = ".safety-project.ini"

MOST_FREQUENTLY_DOWNLOADED_PYPI_PACKAGES = [
    "boto3",
    "urllib3",
    "botocore",
    "requests",
    "setuptools",
    "certifi",
    "idna",
    "charset-normalizer",
    "aiobotocore",
    "typing-extensions",
    "python-dateutil",
    "s3transfer",
    "packaging",
    "grpcio-status",
    "s3fs",
    "six",
    "fsspec",
    "pyyaml",
    "numpy",
    "importlib-metadata",
    "cryptography",
    "zipp",
    "cffi",
    "pip",
    "pandas",
    "google-api-core",
    "pycparser",
    "pydantic",
    "protobuf",
    "wheel",
    "jmespath",
    "attrs",
    "rsa",
    "pyasn1",
    "click",
    "platformdirs",
    "pytz",
    "colorama",
    "jinja2",
    "awscli",
    "markupsafe",
    "tomli",
    "pyjwt",
    "googleapis-common-protos",
    "filelock",
    "virtualenv",
    "cachetools",
    "wrapt",
    "google-auth",
    "pluggy",
    "pytest",
    "pydantic-core",
    "pyparsing",
    "docutils",
    "pyarrow",
    "pyasn1-modules",
    "requests-oauthlib",
    "aiohttp",
    "scipy",
    "jsonschema",
    "oauthlib",
    "sqlalchemy",
    "iniconfig",
    "exceptiongroup",
    "yarl",
    "decorator",
    "multidict",
    "psutil",
    "soupsieve",
    "greenlet",
    "tzdata",
    "pillow",
    "isodate",
    "pygments",
    "beautifulsoup4",
    "annotated-types",
    "requests-toolbelt",
    "frozenlist",
    "tomlkit",
    "pyopenssl",
    "aiosignal",
    "distlib",
    "async-timeout",
    "more-itertools",
    "openpyxl",
    "tqdm",
    "et-xmlfile",
    "grpcio",
    "deprecated",
    "cloudpickle",
    "lxml",
    "pynacl",
    "werkzeug",
    "proto-plus",
    "azure-core",
    "google-cloud-storage",
    "asn1crypto",
    "coverage",
    "websocket-client",
    "msgpack",
    "h11",
    "rich",
    "dill",
    "pexpect",
    "sniffio",
    "anyio",
    "mypy-extensions",
    "ptyprocess",
    "importlib-resources",
    "sortedcontainers",
    "matplotlib",
    "chardet",
    "rpds-py",
    "grpcio-tools",
    "aiohappyeyeballs",
    "flask",
    "httpx",
    "referencing",
    "scikit-learn",
    "jsonschema-specifications",
    "httpcore",
    "pyzmq",
    "poetry-core",
    "keyring",
    "google-cloud-core",
    "python-dotenv",
    "pathspec",
    "markdown-it-py",
    "pkginfo",
    "msal",
    "networkx",
    "bcrypt",
    "mdurl",
    "gitpython",
    "psycopg2-binary",
    "poetry-plugin-export",
    "google-resumable-media",
    "paramiko",
    "kiwisolver",
    "smmap",
    "gitdb",
    "xmltodict",
    "snowflake-connector-python",
    "tabulate",
    "cycler",
    "typedload",
    "jaraco-classes",
    "jeepney",
    "secretstorage",
    "ruamel-yaml",
    "tenacity",
    "wcwidth",
    "build",
    "backoff",
    "shellingham",
    "threadpoolctl",
    "regex",
    "itsdangerous",
    "portalocker",
    "py",
    "google-crc32c",
    "rapidfuzz",
    "pyproject-hooks",
    "py4j",
    "google-cloud-bigquery",
    "fastjsonschema",
    "sqlparse",
    "mccabe",
    "pytest-cov",
    "awswrangler",
    "trove-classifiers",
    "msal-extensions",
    "azure-storage-blob",
    "google-api-python-client",
    "pycodestyle",
    "joblib",
    "google-auth-oauthlib",
    "ruamel-yaml-clib",
    "tzlocal",
    "docker",
    "alembic",
    "fonttools",
    "prompt-toolkit",
    "cachecontrol",
    "azure-identity",
    "distro",
    "marshmallow",
    "uritemplate",
    "isort",
    "cython",
    "ply",
    "httplib2",
    "redis",
    "pymysql",
    "pyrsistent",
    "gym-notices",
    "google-auth-httplib2",
    "poetry",
    "blinker",
    "defusedxml",
    "dnspython",
    "dulwich",
    "toml",
    "gunicorn",
    "crashtest",
    "markdown",
    "nest-asyncio",
    "babel",
    "cleo",
    "sentry-sdk",
    "opentelemetry-api",
    "scramp",
    "multiprocess",
    "installer",
    "termcolor",
    "black",
    "huggingface-hub",
    "mock",
    "msrest",
    "pendulum",
    "requests-aws4auth",
    "ipython",
    "pyflakes",
    "pycryptodomex",
    "grpc-google-iam-v1",
    "types-requests",
    "azure-common",
    "traitlets",
    "fastapi",
    "setuptools-scm",
    "tornado",
    "flake8",
    "contourpy",
    "prometheus-client",
    "future",
    "openai",
    "mako",
    "pycryptodome",
    "imageio",
    "jedi",
    "webencodings",
    "pygithub",
    "parso",
    "transformers",
    "typing-inspect",
    "kubernetes",
    "jsonpointer",
    "matplotlib-inline",
    "starlette",
    "loguru",
    "opentelemetry-sdk",
    "retry",
    "argcomplete",
    "pkgutil-resolve-name",
    "redshift-connector",
    "elasticsearch",
    "pymongo",
    "opentelemetry-semantic-conventions",
    "pytzdata",
    "pytest-runner",
    "asgiref",
    "pg8000",
    "bs4",
    "datadog",
    "debugpy",
    "python-json-logger",
    "jsonpath-ng",
    "uvicorn",
    "executing",
    "smart-open",
    "zope-interface",
    "asttokens",
    "typer",
    "aioitertools",
    "apache-airflow",
    "sagemaker",
    "arrow",
    "google-pasta",
    "pyspark",
    "humanfriendly",
    "websockets",
    "stack-data",
    "shapely",
    "pure-eval",
    "torch",
    "oscrypto",
    "tokenizers",
    "pysocks",
    "sphinx",
    "typeguard",
    "tox",
    "scikit-image",
    "requests-file",
    "google-cloud-pubsub",
    "pytest-mock",
    "google-cloud-secret-manager",
    "snowflake-sqlalchemy",
    "mysql-connector-python",
    "pylint",
    "jupyter-core",
    "jupyter-client",
    "astroid",
    "jsonpatch",
    "setproctitle",
    "adal",
    "types-python-dateutil",
    "ipykernel",
    "xgboost",
    "orjson",
    "schema",
    "tb-nightly",
    "nbconvert",
    "xlrd",
    "toolz",
    "appdirs",
    "aiofiles",
    "sympy",
    "opensearch-py",
    "nodeenv",
    "pywavelets",
    "jaraco-functools",
    "jupyter-server",
    "nbformat",
    "jupyterlab",
    "progressbar2",
    "comm",
    "identify",
    "bleach",
    "mypy",
    "pathos",
    "pyodbc",
    "pre-commit",
    "xlsxwriter",
    "rfc3339-validator",
    "aws-requests-auth",
    "gym",
    "pox",
    "ppft",
    "mistune",
    "aenum",
    "jaraco-context",
    "tinycss2",
    "pbr",
    "google-cloud-appengine-logging",
    "notebook",
    "db-dtypes",
    "mpmath",
    "sentencepiece",
    "responses",
    "cfgv",
    "cattrs",
    "python-utils",
    "slack-sdk",
    "jupyterlab-server",
    "nbclient",
    "lz4",
    "ipywidgets",
    "sshtunnel",
    "absl-py",
    "widgetsnbextension",
    "watchdog",
    "asynctest",
    "semver",
    "rfc3986",
    "google-cloud-aiplatform",
    "jupyterlab-widgets",
    "altair",
    "pandas-gbq",
    "click-man",
    "tensorboard",
    "smdebug-rulesconfig",
    "simplejson",
    "text-unidecode",
    "argon2-cffi",
    "apache-airflow-providers-common-sql",
    "snowballstemmer",
    "azure-mgmt-core",
    "docker-pycreds",
    "nltk",
    "python-slugify",
    "croniter",
    "structlog",
    "selenium",
    "antlr4-python3-runtime",
    "google-cloud-logging",
    "argon2-cffi-bindings",
    "azure-storage-file-datalake",
    "django",
    "pydeequ",
    "pytest-xdist",
    "h5py",
    "google-cloud-resource-manager",
    "dataclasses",
    "execnet",
    "send2trash",
    "opentelemetry-proto",
    "google-cloud-bigquery-storage",
    "oauth2client",
    "dataclasses-json",
    "json5",
    "tiktoken",
    "wandb",
    "databricks-sql-connector",
    "langchain-core",
    "overrides",
    "prettytable",
    "pandocfilters",
    "semantic-version",
    "jupyterlab-pygments",
    "msrestazure",
    "safetensors",
    "hvac",
    "colorlog",
    "imbalanced-learn",
    "monotonic",
    "seaborn",
    "alabaster",
    "terminado",
    "webcolors",
    "ordered-set",
    "graphql-core",
    "notebook-shim",
    "lazy-object-proxy",
    "funcsigs",
    "numba",
    "llvmlite",
    "gremlinpython",
    "xxhash",
    "great-expectations",
    "flatbuffers",
    "pydata-google-auth",
    "fqdn",
    "uri-template",
    "imagesize",
    "opentelemetry-exporter-otlp-proto-common",
    "isoduration",
    "backports-tarfile",
    "wsproto",
    "tensorflow",
    "thrift",
    "hypothesis",
    "rfc3986-validator",
    "trio",
    "inflection",
    "html5lib",
    "plotly",
    "entrypoints",
    "sphinxcontrib-serializinghtml",
    "jupyter-events",
    "lockfile",
    "coloredlogs",
    "sphinxcontrib-htmlhelp",
    "cached-property",
    "sphinxcontrib-qthelp",
    "sphinxcontrib-devhelp",
    "sphinxcontrib-applehelp",
    "gast",
    "azure-cli",
    "azure-datalake-store",
    "opentelemetry-exporter-otlp-proto-http",
    "pyproject-api",
    "azure-mgmt-resource",
    "async-lru",
    "faker",
    "sphinxcontrib-jsmath",
    "nose",
    "opencv-python",
    "outcome",
    "statsmodels",
    "readme-renderer",
    "jupyter-server-terminals",
    "libcst",
    "retrying",
    "datasets",
    "aniso8601",
    "pybind11",
    "databricks-sdk",
    "pyroaring",
    "azure-keyvault-secrets",
    "email-validator",
    "argparse",
    "parameterized",
    "docopt",
    "google-cloud-audit-log",
    "confluent-kafka",
    "kafka-python",
    "pymssql",
    "zeep",
    "gcsfs",
    "click-plugins",
    "jupyter-lsp",
    "ruff",
    "deepdiff",
    "docstring-parser",
    "tblib",
    "time-machine",
    "jiter",
    "patsy",
    "azure-storage-common",
    "deprecation",
    "azure-nspkg",
    "databricks-cli",
    "nh3",
    "twine",
    "invoke",
    "delta-spark",
    "watchtower",
    "mlflow",
    "pydantic-settings",
    "azure-mgmt-storage",
    "opentelemetry-exporter-otlp-proto-grpc",
    "applicationinsights",
    "dbt-core",
    "freezegun",
    "pickleshare",
    "apache-airflow-providers-ssh",
    "python-multipart",
    "langchain",
    "uv",
    "unidecode",
    "azure-keyvault-keys",
    "azure-cosmos",
    "pytest-metadata",
    "pipenv",
    "tensorboard-data-server",
    "azure-graphrbac",
    "google-cloud-kms",
    "backcall",
    "trio-websocket",
    "azure-keyvault",
    "pytest-asyncio",
    "psycopg2",
    "google-cloud-dataproc",
    "keras",
    "datetime",
    "zope-event",
    "apache-airflow-providers-google",
    "backports-zoneinfo",
    "google-cloud-monitoring",
    "looker-sdk",
    "azure-mgmt-containerregistry",
    "makefun",
    "google-cloud-vision",
    "mlflow-skinny",
    "hatchling",
    "spacy",
    "torchvision",
    "apache-airflow-providers-snowflake",
    "google-cloud-spanner",
    "google-cloud-container",
    "nvidia-nccl-cu12",
    "triton",
    "gevent",
    "google-cloud-dlp",
    "uvloop",
    "simple-salesforce",
    "tldextract",
    "analytics-python",
    "apache-airflow-providers-databricks",
    "tensorflow-estimator",
    "google-cloud-bigquery-datatransfer",
    "azure-mgmt-keyvault",
    "azure-mgmt-cosmosdb",
    "azure-mgmt-compute",
    "graphviz",
    "google-cloud-tasks",
    "ujson",
    "opentelemetry-instrumentation",
    "azure-mgmt-authorization",
    "fastavro",
    "httptools",
    "pathlib2",
    "azure-mgmt-network",
    "google-cloud-datacatalog",
    "pkce",
    "google-ads",
    "opt-einsum",
    "sh",
    "jsondiff",
    "azure-mgmt-msi",
    "google-cloud-firestore",
    "evergreen-py",
    "google-cloud-bigtable",
    "astunparse",
    "watchfiles",
    "configparser",
    "flask-appbuilder",
    "fabric",
    "azure-mgmt-recoveryservices",
    "apache-airflow-providers-mysql",
    "scp",
    "db-contrib-tool",
    "google-cloud-build",
    "omegaconf",
    "azure-mgmt-monitor",
    "ecdsa",
    "gspread",
    "azure-mgmt-signalr",
    "azure-mgmt-containerinstance",
    "blis",
    "thinc",
    "bitarray",
    "murmurhash",
    "pycrypto",
    "dask",
    "requests-mock",
    "catalogue",
    "cymem",
    "azure-mgmt-sql",
    "preshed",
    "google-cloud-workflows",
    "opentelemetry-exporter-otlp",
    "azure-mgmt-web",
    "google-cloud-redis",
    "azure-batch",
    "kombu",
    "pywin32",
    "azure-data-tables",
    "wasabi",
    "azure-mgmt-containerservice",
    "azure-mgmt-servicebus",
    "azure-mgmt-redis",
    "google-cloud-dataplex",
    "srsly",
    "pytimeparse",
    "google-cloud-language",
    "authlib",
    "google-cloud-automl",
    "google-cloud-videointelligence",
    "google-cloud-os-login",
    "azure-mgmt-rdbms",
    "brotli",
    "pyserial",
    "azure-mgmt-dns",
    "langchain-community",
    "nvidia-cudnn-cu12",
    "texttable",
    "azure-mgmt-advisor",
    "google-cloud-memcache",
    "azure-mgmt-eventhub",
    "tensorflow-serving-api",
    "gsutil",
    "lark",
    "azure-cli-core",
    "flask-cors",
    "pysftp",
    "celery",
    "langcodes",
    "azure-mgmt-batch",
    "azure-mgmt-loganalytics",
    "azure-mgmt-cdn",
    "ninja",
    "azure-mgmt-recoveryservicesbackup",
    "azure-mgmt-iothub",
    "azure-mgmt-search",
    "azure-mgmt-marketplaceordering",
    "azure-mgmt-trafficmanager",
    "azure-mgmt-managementgroups",
    "pip-tools",
    "azure-mgmt-cognitiveservices",
    "azure-mgmt-devtestlabs",
    "azure-mgmt-eventgrid",
    "python-gnupg",
    "jira",
    "pypdf2",
    "azure-mgmt-applicationinsights",
    "azure-mgmt-servicefabric",
    "billiard",
    "azure-mgmt-media",
    "azure-mgmt-billing",
    "ratelimit",
    "azure-mgmt-iothubprovisioningservices",
    "azure-mgmt-policyinsights",
    "azure-mgmt-nspkg",
    "google-cloud-orchestration-airflow",
    "apache-airflow-providers-cncf-kubernetes",
    "azure-mgmt-batchai",
    "azure-mgmt-iotcentral",
    "azure-mgmt-datamigration",
    "graphql-relay",
    "azure-mgmt-maps",
    "graphene",
    "azure-appconfiguration",
    "amqp",
    "google-cloud-dataproc-metastore",
    "mdit-py-plugins",
    "google-cloud-translate",
    "ijson",
    "sqlalchemy-bigquery",
    "vine",
    "nvidia-cublas-cu12",
    "nvidia-nvjitlink-cu12",
    "spacy-loggers",
    "spacy-legacy",
    "levenshtein",
    "agate",
    "azure-mgmt-datalake-nspkg",
    "knack",
    "yapf",
    "awscrt",
    "azure-mgmt-datalake-store",
    "google-cloud-dataform",
    "types-pyyaml",
    "confection",
    "propcache",
    "google-cloud-speech",
    "nvidia-cuda-runtime-cu12",
    "opencensus",
    "opencensus-context",
    "nvidia-cuda-cupti-cu12",
    "nvidia-cuda-nvrtc-cu12",
    "parsedatetime",
    "nvidia-cusparse-cu12",
    "nvidia-cufft-cu12",
    "nvidia-cusolver-cu12",
    "grpcio-gcp",
    "nvidia-curand-cu12",
    "google-cloud-texttospeech",
    "typing",
    "humanize",
    "pytest-html",
    "langsmith",
    "nvidia-nvtx-cu12",
    "flask-sqlalchemy",
    "opentelemetry-util-http",
    "narwhals",
    "azure-multiapi-storage",
    "gcloud-aio-storage",
    "pycountry",
    "jsonpickle",
    "zstandard",
    "avro-python3",
    "libclang",
    "apispec",
    "gcloud-aio-auth",
    "azure-storage-queue",
    "contextlib2",
    "azure-mgmt-datalake-analytics",
    "gcloud-aio-bigquery",
    "azure-mgmt-reservations",
    "javaproperties",
    "tensorflow-io-gcs-filesystem",
    "azure-loganalytics",
    "djangorestframework",
    "azure-mgmt-consumption",
    "hpack",
    "google-cloud-compute",
    "click-didyoumean",
    "azure-mgmt-relay",
    "parsimonious",
    "azure-synapse-artifacts",
    "python-magic",
    "azure-cli-telemetry",
    "click-repl",
    "moto",
    "pyathena",
    "pyproj",
    "protobuf3-to-dict",
    "durationpy",
    "stevedore",
    "python-daemon",
    "azure-synapse-spark",
    "apache-airflow-providers-http",
    "mypy-boto3-s3",
    "pyspnego",
    "cfn-lint",
    "astor",
    "azure-mgmt-apimanagement",
    "h2",
    "hyperframe",
    "azure-mgmt-hdinsight",
    "azure-mgmt-privatedns",
    "boto3-stubs",
    "mashumaro",
    "dateparser",
    "ml-dtypes",
    "mysqlclient",
    "azure-mgmt-security",
    "opencensus-ext-azure",
    "azure-mgmt-synapse",
    "azure-mgmt-kusto",
    "azure-mgmt-netapp",
    "grpcio-health-checking",
    "azure-mgmt-redhatopenshift",
    "iso8601",
    "lightgbm",
    "azure-mgmt-appconfiguration",
    "azure-keyvault-administration",
    "boto",
    "azure-mgmt-sqlvirtualmachine",
    "azure-mgmt-imagebuilder",
    "azure-synapse-accesscontrol",
    "enum34",
    "azure-mgmt-servicelinker",
    "azure-mgmt-botservice",
    "azure-mgmt-servicefabricmanagedclusters",
    "jpype1",
    "python-jose",
    "azure-mgmt-databoxedge",
    "azure-synapse-managedprivateendpoints",
    "azure-mgmt-extendedlocation",
    "office365-rest-python-client",
    "onnxruntime",
    "azure-mgmt-managedservices",
    "cramjam",
    "urllib3-secure-extra",
    "avro",
    "holidays",
    "psycopg",
    "botocore-stubs",
    "fasteners",
    "resolvelib",
    "partd",
    "hyperlink",
    "leather",
    "apscheduler",
    "flask-wtf",
    "jupyter",
    "marisa-trie",
    "locket",
    "jupyter-console",
    "python-http-client",
    "elastic-transport",
    "dbt-extractor",
    "tensorflow-text",
    "language-data",
    "inflect",
    "fuzzywuzzy",
    "cytoolz",
    "cmake",
    "parse",
    "python-gitlab",
    "mypy-boto3-rds",
    "tifffile",
    "eth-utils",
    "eth-hash",
    "netaddr",
    "incremental",
    "setuptools-rust",
    "python-levenshtein",
    "geopandas",
    "twisted",
    "langchain-text-splitters",
    "types-awscrt",
    "apache-airflow-providers-fab",
    "yamllint",
    "cligj",
    "sphinx-rtd-theme",
    "azure-mgmt-deploymentmanager",
    "pytest-timeout",
    "lazy-loader",
    "wtforms",
    "bytecode",
    "accelerate",
    "polars",
    "sendgrid",
    "frozendict",
    "flask-login",
    "opentelemetry-instrumentation-requests",
    "jaydebeapi",
    "eth-typing",
    "dacite",
    "types-pytz",
    "py-cpuinfo",
    "querystring-parser",
    "universal-pathlib",
    "dbt-semantic-interfaces",
    "magicattr",
    "cssselect",
    "fastparquet",
    "opencv-python-headless",
    "automat",
    "unicodecsv",
    "constantly",
    "kfp",
    "ddtrace",
    "logbook",
    "envier",
    "cloudpathlib",
    "types-s3transfer",
    "google-cloud-dataflow-client",
    "sqlalchemy-utils",
    "apache-beam",
    "validators",
    "bracex",
    "apache-airflow-providers-ftp",
    "phonenumbers",
    "diskcache",
    "mergedeep",
    "slicer",
    "shap",
    "python-docx",
    "types-urllib3",
    "pytest-rerunfailures",
    "types-setuptools",
    "pathy",
    "pytz-deprecation-shim",
    "yappi",
    "pydot",
    "types-protobuf",
    "ipython-genutils",
    "pytorch-lightning",
    "fire",
    "apache-airflow-providers-sqlite",
    "nvidia-cublas-cu11",
    "azure-storage-file-share",
    "mmh3",
    "azure-mgmt-datafactory",
    "azure-servicebus",
    "nvidia-cudnn-cu11",
    "inject",
    "typed-ast",
    "connexion",
    "configargparse",
    "linkify-it-py",
    "aws-sam-translator",
    "slackclient",
    "eth-abi",
    "pydash",
    "timm",
    "datadog-api-client",
    "nvidia-cuda-runtime-cu11",
    "nvidia-cuda-nvrtc-cu11",
    "geographiclib",
    "gradio",
    "cron-descriptor",
    "ansible",
    "azure-kusto-data",
    "django-cors-headers",
    "junit-xml",
    "geopy",
    "uc-micro-py",
    "pyee",
    "xarray",
    "ansible-core",
    "pypdf",
    "pyotp",
    "starkbank-ecdsa",
    "geoip2",
    "multimethod",
    "eth-account",
    "meson",
    "jellyfish",
    "futures",
    "cachelib",
    "flask-caching",
    "natsort",
    "autopep8",
    "torchaudio",
    "torchmetrics",
    "pydub",
    "pandera",
    "pyhcl",
    "apache-airflow-providers-slack",
    "oracledb",
    "google-cloud-run",
    "h3",
    "apache-airflow-providers-amazon",
    "sqlalchemy-spanner",
    "events",
    "google-cloud-batch",
    "requests-ntlm",
    "bottle",
    "google-cloud-storage-transfer",
    "junitparser",
    "apache-airflow-providers-smtp",
    "apache-airflow-providers-imap",
    "emoji",
    "crcmod",
    "statsd",
    "limits",
    "apache-airflow-providers-common-io",
    "methodtools",
    "asyncpg",
    "strictyaml",
    "wcmatch",
    "marshmallow-sqlalchemy",
    "faiss-cpu",
    "sentence-transformers",
    "psycopg-binary",
    "azure-keyvault-certificates",
    "django-filter",
    "maxminddb",
    "weasel",
    "gql",
    "onnx",
    "fiona",
    "boltons",
    "dbt-common",
    "bidict",
    "keras-applications",
    "json-merge-patch",
    "elasticsearch-dsl",
    "ftfy",
    "swagger-ui-bundle",
    "tableauserverclient",
    "flask-jwt-extended",
    "lightning-utilities",
    "meson-python",
    "google-cloud",
]



import abc
import json
import os.path
import re
import subprocess
from pathlib import Path
from sys import platform
from tempfile import mkstemp
import time

import nltk
import typer
from filelock import FileLock
from rich.padding import Padding
from rich.prompt import Prompt

from safety.console import main_console as console
from safety.constants import PIP_LOCK
from safety.models import ToolResult
from safety.tool.constants import (
    MOST_FREQUENTLY_DOWNLOADED_PYPI_PACKAGES,
    PROJECT_CONFIG,
)
from safety.tool.pip import Pip
from safety.tool.poetry import Poetry
from safety.tool.resolver import get_unwrapped_command
from safety.tool.environment_diff import PipEnvironmentDiffTracker

from safety.events.utils import emit_diff_operations, emit_tool_command_executed

from safety_schemas.models.events.types import ToolType

from typing_extensions import List

from typing import Any, Dict, TYPE_CHECKING, Optional

if TYPE_CHECKING:
    pass


def is_os_supported():
    return platform in ["linux", "linux2", "darwin", "win32"]


class BuildFileConfigurator(abc.ABC):
    @abc.abstractmethod
    def is_supported(self, file: Path) -> bool:
        """
        Returns whether a specific file is supported by this class.
        Args:
            file (str): The file to check.
        Returns:
            bool: Whether the file is supported by this class.
        """
        pass

    @abc.abstractmethod
    def configure(
        self, file: Path, org_slug: Optional[str], project_id: Optional[str]
    ) -> None:
        """
        Configures specific file.
        Args:
            file (str): The file to configure.
            org_slug (str): The organization slug.
            project_id (str): The project id.
        """
        pass


class PipRequirementsConfigurator(BuildFileConfigurator):
    __file_name_pattern = re.compile("^([a-zA-Z_-]+)?requirements([a-zA-Z_-]+)?.txt$")

    def is_supported(self, file: Path) -> bool:
        return self.__file_name_pattern.match(os.path.basename(file)) is not None

    def configure(
        self, file: Path, org_slug: Optional[str], project_id: Optional[str]
    ) -> None:
        Pip.configure_requirements(file, org_slug, project_id)  # type: ignore


class PoetryPyprojectConfigurator(BuildFileConfigurator):
    __file_name_pattern = re.compile("^pyproject.toml$")

    def is_supported(self, file: Path) -> bool:
        return self.__file_name_pattern.match(
            os.path.basename(file)
        ) is not None and Poetry.is_poetry_project_file(file)

    def configure(
        self, file: Path, org_slug: Optional[str], project_id: Optional[str]
    ) -> None:
        Poetry.configure_pyproject(file, org_slug, project_id)  # type: ignore


# TODO: Review if we should move this/hook up this into interceptors.
class ToolConfigurator(abc.ABC):
    @abc.abstractmethod
    def configure(self, org_slug: Optional[str]) -> None:
        """
        Configures specific tool.
        Args:
            org_slug (str): The organization slug.
        """
        pass

    @abc.abstractmethod
    def reset(self) -> None:
        """
        Resets specific tool.
        """
        pass


class PipConfigurator(ToolConfigurator):
    def configure(self, org_slug: Optional[str]) -> None:
        Pip.configure_system(org_slug)

    def reset(self) -> None:
        Pip.reset_system()


class BaseCommand(abc.ABC):
    @abc.abstractmethod
    def execute(self, ctx: typer.Context) -> None:
        """
        Executes the command.
        Args:
            ctx (typer.Context): The context.
        """
        pass

    @abc.abstractmethod
    def before(self, ctx: typer.Context) -> None:
        """
        Executes before the command.
        Args:
            ctx (typer.Context): The context.
        """
        pass

    @abc.abstractmethod
    def after(self, ctx: typer.Context, result: ToolResult) -> None:
        """
        Executes after the command.
        Args:
            ctx (typer.Context): The context.
            result (ToolResult): The result.
        """
        pass

    @abc.abstractmethod
    def env(self, ctx: typer.Context) -> dict:
        """
        Returns the environment.
        Args:
            ctx (typer.Context): The context.
        Returns:
            dict: The environment.
        """
        pass


class PipCommand(BaseCommand):
    def __init__(self, args: List[str], capture_output: bool = False) -> None:
        self._args = args
        self.__capture_output = capture_output
        self.__filelock = FileLock(PIP_LOCK, 10)
        self._diff_tracker = PipEnvironmentDiffTracker()
        self._name = ["pip"]

    def _initialize_diff_tracker(self, ctx: typer.Context):
        """
        Common implementation to initialize the diff tracker.
        Can be called by child classes in their before() implementation.
        """
        current_packages = self._get_installed_packages(ctx)
        self._diff_tracker.set_before_state(current_packages)

    def _handle_command_result(self, ctx: typer.Context, result: ToolResult):
        """
        Common implementation to handle command results.
        Can be called by child classes in their after() implementation.
        """
        process = result.process
        if process:
            if process.returncode == 0:
                self._perform_diff(ctx)

            emit_tool_command_executed(
                ctx.obj.event_bus,
                ctx,  # type: ignore
                tool=ToolType.PIP,
                result=result,
            )

    def before(self, ctx: typer.Context):
        self._initialize_diff_tracker(ctx)

    def after(self, ctx: typer.Context, result: ToolResult):
        self._handle_command_result(ctx, result)

    def execute(self, ctx: typer.Context):
        with self.__filelock:
            self.before(ctx)
            # TODO: Safety should redirect to the proper pip, if the user is
            # using pip3, it should be redirected to pip3, not pip to avoid any
            # issues.

            pre_args = [get_unwrapped_command(name=self._name[0])]
            args = pre_args + self.__remove_safety_args(self._args)

            started_at = time.monotonic()
            process = subprocess.run(
                args, capture_output=self.__capture_output, env=self.env(ctx)
            )

            duration_ms = int((time.monotonic() - started_at) * 1000)

            result = ToolResult(process=process, duration_ms=duration_ms)

            self.after(ctx, result)

    def env(self, ctx: typer.Context):
        return os.environ.copy()

    @classmethod
    def from_args(cls, args: List[str]):
        if "install" in args:
            return PipInstallCommand(args)
        elif "uninstall" in args:
            return PipUninstallCommand(args)
        else:
            return PipGenericCommand(args)

    def __remove_safety_args(self, args: List[str]):
        return [arg for arg in args if not arg.startswith("--safety")]

    def _get_installed_packages(self, ctx: typer.Context) -> List[Dict[str, Any]]:
        """
        Get the currently installed packages as a Python dictionary.
        """
        pre_args = [get_unwrapped_command(name=self._name[0])] + self._name[1:]
        args = pre_args + ["list", "--format=json"]
        result = subprocess.run(args, capture_output=True, env=self.env(ctx), text=True)
        # TODO: Handle error
        return json.loads(result.stdout)

    def _perform_diff(self, ctx: typer.Context):
        """
        Perform the diff operation.
        Can be called by child classes when appropriate.
        """
        current_packages = self._get_installed_packages(ctx)
        self._diff_tracker.set_after_state(current_packages)
        added, removed, updated = self._diff_tracker.get_diff()

        emit_diff_operations(
            ctx.obj.event_bus,
            ctx,  # type: ignore
            added=added,
            removed=removed,
            updated=updated,
            by_tool=ToolType.PIP,
        )


class PipGenericCommand(PipCommand):
    pass


class PipInstallCommand(PipCommand):
    def __init__(self, args: List[str]) -> None:
        super().__init__(args)
        self.package_names = []
        self.__index_url = None

    def before(self, ctx: typer.Context):
        super().before(ctx)
        args = self._args

        ranges_to_delete = []
        for ind, val in enumerate(args):
            if ind > 0 and (
                args[ind - 1].startswith("-i")
                or args[ind - 1].startswith("--index-url")
            ):
                if args[ind].startswith("https://pkgs.safetycli.com"):
                    self.__index_url = args[ind]

                ranges_to_delete.append((ind - 1, ind))
            elif ind > 0 and (
                args[ind - 1] == "-r" or args[ind - 1] == "--requirement"
            ):
                requirement_file = args[ind]

                if not Path(requirement_file).is_file():
                    continue

                with open(requirement_file, "r") as f:
                    fd, tmp_requirements_path = mkstemp(
                        suffix="safety-requirements.txt", text=True
                    )
                    with os.fdopen(fd, "w") as tf:
                        requirements = re.sub(
                            r"^(-i|--index-url).*$", "", f.read(), flags=re.MULTILINE
                        )
                        tf.write(requirements)

                    args[ind] = tmp_requirements_path
            elif (
                ind > 0
                and (
                    not args[ind - 1].startswith("-e")
                    or not args[ind - 1].startswith("--editable")
                )
                and not args[ind].startswith("-")
            ):
                if args[ind] == ".":
                    continue

                package_name = args[ind]
                (valid, candidate_package_name) = self.__check_typosquatting(
                    package_name
                )
                if not valid:
                    prompt = f"You are about to install {package_name} package. Did you mean to install {candidate_package_name}?"
                    answer = Prompt.ask(
                        prompt=prompt,
                        choices=["y", "n"],
                        default="y",
                        show_default=True,
                        console=console,
                    ).lower()
                    if answer == "y":
                        package_name = candidate_package_name
                        console.print(f"Installing {package_name} package instead.")
                        args[ind] = package_name

                self.__add_package_name(package_name)

        for start, end in ranges_to_delete:
            args = args[:start] + args[end + 1 :]

        self._args = args

    def after(self, ctx: typer.Context, result: ToolResult):
        super().after(ctx, result)

        if result.process and result.process.returncode == 0:
            self.__run_scan()
        else:
            self.__render_package_details()

    def env(self, ctx: typer.Context) -> dict:
        env = super().env(ctx)
        env["PIP_INDEX_URL"] = (
            Pip.build_index_url(ctx, self.__index_url)
            if not self.__is_check_disabled()
            else Pip.default_index_url()
        )
        return env

    def __is_check_disabled(self):
        return "--safety-disable-check" in self._args

    def __check_typosquatting(self, package_name):
        max_edit_distance = 2 if len(package_name) > 5 else 1

        if package_name in MOST_FREQUENTLY_DOWNLOADED_PYPI_PACKAGES:
            return (True, package_name)

        for pkg in MOST_FREQUENTLY_DOWNLOADED_PYPI_PACKAGES:
            if (
                abs(len(pkg) - len(package_name)) <= max_edit_distance
                and nltk.edit_distance(pkg, package_name) <= max_edit_distance
            ):
                return (False, pkg)

        return (True, package_name)

    def __run_scan(self):
        if not is_os_supported():
            return

        target = os.getcwd()
        if Path(os.path.join(target, PROJECT_CONFIG)).is_file():
            try:
                subprocess.Popen(
                    ["safety", "scan"],
                    stdout=subprocess.DEVNULL,
                    stderr=subprocess.DEVNULL,
                    stdin=subprocess.DEVNULL,
                    start_new_session=True,
                )
            except Exception:
                pass

    def __add_package_name(self, package_name):
        r = re.compile(r"^([a-zA-Z_-]+)(([~<>=]=)[a-zA-Z0-9._-]+)?")
        match = r.match(package_name)
        if match:
            self.package_names.append(match.group(1))

    def __render_package_details(self):
        for package_name in self.package_names:
            console.print(
                Padding(
                    f"Learn more: [link]https://data.safetycli.com/packages/pypi/{package_name}/[/link]",
                    (0, 0, 0, 1),
                ),
                emoji=True,
            )


class PipUninstallCommand(PipCommand):
    pass


class UVBaseCommand(PipCommand):
    def __init__(self, args: List[str]) -> None:
        super().__init__(args)
        self._name = ["uv", "pip"]


class UvInstallCommand(PipInstallCommand, UVBaseCommand):
    def before(self, ctx: typer.Context):
        pass


class UvUninstallCommand(PipUninstallCommand, UVBaseCommand):
    pass


class UvGenericCommand(PipGenericCommand, UVBaseCommand):
    pass


class UvCommand:
    @classmethod
    def from_args(cls, args: List[str]):
        if "install" in args:
            return UvInstallCommand(args)
        elif "uninstall" in args:
            return UvUninstallCommand(args)
        else:
            return UvGenericCommand(args)



from safety.meta import get_version

# Console Help Theme
CONSOLE_HELP_THEME = {
    "nhc": "grey82"
}

CLI_VERSION = get_version()
CLI_WEBSITE_URL="https://safetycli.com"
CLI_DOCUMENTATION_URL="https://docs.safetycli.com"
CLI_SUPPORT_EMAIL="support@safetycli.com"

# Main Safety --help data:
CLI_MAIN_INTRODUCTION = f"Safety CLI 3 - Vulnerability Scanning for Secure Python Development\n\n" \
"Leverage the most comprehensive vulnerability data available to secure your projects against vulnerable and malicious packages. Safety CLI is a Python dependency vulnerability scanner that enhances software supply chain security at every stage of development.\n\n" \
f"Documentation: {CLI_DOCUMENTATION_URL}\n"\
f"Contact: {CLI_SUPPORT_EMAIL}\n"

CLI_AUTH_COMMAND_HELP = (
    "Authenticate Safety CLI to perform scans. Your default browser will automatically open to "
    "https://platform.safetycli.com.\n\n"
    "Example:\n  safety auth login\n\n"
    "For headless authentication, you will receive a URL to paste into an external browser.\n\n"
    "Example:\n  safety auth login --headless"
)

CLI_AUTH_HEADLESS_HELP = "For headless authentication, you will receive a URL to paste into an external browser."

CLI_SCAN_COMMAND_HELP = "Scans a Python project directory."\
"\nExample: safety scan to scan the current directory"
CLI_SYSTEM_SCAN_COMMAND_HELP = "\\[beta] Run a comprehensive scan for packages and vulnerabilities across your entire machine/environment."\
"\nExample: safety system-scan"

CLI_CHECK_COMMAND_HELP = "\\[deprecated] Find vulnerabilities at target files or environments. Now replaced by safety scan, and will be unsupported beyond 1 May 2024." \
"\nExample: safety check -r requirements.txt"
CLI_LICENSES_COMMAND_HELP = "\\[deprecated] Find licenses at target files or environments. This command will be replaced by safety scan, and will be unsupported beyond 1 May 2024." \
"\nExample: safety license -r requirements.txt"


CLI_ALERT_COMMAND_HELP = "\\[deprecated] Create GitHub pull requests or GitHub issues using a `safety check` json report file. Being replaced by newer features." \
"\nExample: safety alert --check-report your-report.json --key API_KEY github-pr --repo my-org/my-repo --token github-token"

CLI_CHECK_UPDATES_HELP = "Check for version updates to Safety CLI."\
"\nExample: safety check-updates"

CLI_CONFIGURE_HELP = "Set up global configurations for Safety CLI, including proxy settings and organization details."\
"\nExample: safety configure --proxy-host 192.168.0.1"

CLI_GENERATE_HELP = "Generate a boilerplate Safety CLI policy file for customized security policies."\
"\nNote: Safety Platform policies will override any local policy files found"\
"\nExample: safety generate policy_file"

CLI_VALIDATE_HELP = "Check if your local Safety CLI policy file is valid."\
"\nExample: Example: safety validate --path /path/to/policy.yml"

CLI_GATEWAY_CONFIGURE_COMMAND_HELP = "Configures the project in the working directory to use Gateway."

# Global options help
_CLI_PROXY_TIP_HELP = f"[nhc]Note: proxy details can be set globally in a config file.[/nhc]\n\nSee [bold]safety configure --help[/bold]\n\n"

CLI_PROXY_HOST_HELP = "Specify a proxy host for network communications. \n\n" + \
 _CLI_PROXY_TIP_HELP

CLI_PROXY_PORT_HELP = "Set the proxy port (default: 80).\n\n" + \
_CLI_PROXY_TIP_HELP

CLI_PROXY_PROTOCOL_HELP = "Choose the proxy protocol (default: https).\n\n" + \
_CLI_PROXY_TIP_HELP

CLI_KEY_HELP = "The API key required for cicd stage or production stage scans.\n\n" \
"[nhc]For development stage scans unset the API key and authenticate using [bold]safety auth[/bold].[/nhc]\n\n" \
"[nhc]Tip: the API key can also be set using the environment variable: SAFETY_API_KEY[/nhc]\n\n"\
"[bold]Example: safety --key API_KEY scan[/bold]"

CLI_STAGE_HELP = "Assign a development lifecycle stage to your scan (default: development).\n\n" \
"[nhc]This labels the scan and its findings in Safety Platform with this stage.[/nhc]\n\n" \
"[bold]Example: safety --stage production scan[/bold]"

CLI_DEBUG_HELP = "Enable debug mode for detailed output.\n\n" \
"[bold]Example: safety --debug scan[/bold]"

CLI_DISABLE_OPTIONAL_TELEMETRY_DATA_HELP = "Opt-out of sending optional telemetry data. Anonymized telemetry data will remain.\n\n" \
"[bold]Example: safety --disable-optional-telemetry scan[/bold]"

# Scan Help options
SCAN_POLICY_FILE_HELP = "Use a local policy file to configure the scan.\n\n" \
    "[nhc]Note: Project scan policies defined in Safety Platform will override local policy files[/nhc]\n\n" \
    "[bold]Example: safety scan --policy-file /path/to/policy.yml[/bold]"
SCAN_TARGET_HELP = "Define a specific project path to scan. (default: current directory)\n\n" \
    "[bold]Example: safety scan --target /path/to/project[/bold]"
SCAN_OUTPUT_HELP = "Set the output format for scan results (default: screen)\n\n" \
    "[bold]Example: safety scan --output json[/bold]"
SCAN_SAVE_AS_HELP = "In addition to regular output save the scan results to a json, html, text, or spdx file using: FORMAT FILE_PATH\n\n" \
    "[bold]Example: safety scan --save-as json results.json[/bold]"
SCAN_DETAILED_OUTPUT = "Enable a verbose scan report for detailed insights (only for screen output)\n\n" \
    "[bold]Example: safety scan --detailed-output[/bold]"
SCAN_APPLY_FIXES = "[bold]Update packages listed in requirements.txt files to secure versions where possible[/bold]\n\n"\
    "[nhc]Currently supports: requirements.txt files[/nhc]\n\n"\
    "Note: this will update your requirements.txt file "

# System Scan options
SYSTEM_SCAN_POLICY_FILE_HELP = "Use a local policy file to configure the scan.\n\n" \
    "[nhc]Note: Scan policies defined in Safety Platform will override local policy files[/nhc]\n\n" \
    "[bold]Example: safety scan --policy-file /path/to/policy.yml[/bold]"
SYSTEM_SCAN_TARGET_HELP = "Define a specific location to start the system scan. (default: current directory)\n\n" \
    "[bold]Example: safety scan --target /path/to/project[/bold]"
SYSTEM_SCAN_OUTPUT_HELP = "Set the output format for scan results (default: screen)\n\n" \
    "[bold]Example: safety scan --output json[/bold]"
SYSTEM_SCAN_SAVE_AS_HELP = "In addition to the terminal/console output (set by --output), save system-scan results to a screen (text) or json file.\n\n" \
    """[nhc]Use [bold]--save-as <FORMAT> <PATH>[/bold]. For example: [bold]--save-as json my-machine-scan.json[/bold] to save the system-scan results to `my-machine-scan.json` in the current directory[/nhc]\n\n""" \
    "[nhc][Default: json .][/nhc]"

# Auth options
CLI_AUTH_LOGIN_HELP = "Authenticate with Safety CLI to perform scans. Your default browser will automatically open to https://platform.safetycli.com unless already authenticated.\n\n" \
    "[bold]Example: safety auth login[/bold]"
CLI_AUTH_LOGOUT_HELP = "Log out from the current Safety CLI session.\n\n" \
    "[bold]Example: safety auth logout[/bold]"
CLI_AUTH_STATUS_HELP = "Show the current authentication status.\n\n" \
    "[bold]Example: safety auth status[/bold]"

DEFAULT_EPILOG = f"\nSafety CLI version: {CLI_VERSION}\n" \
    f"\nDocumentation: {CLI_DOCUMENTATION_URL}\n\n\n\n" \
    "Made with love by Safety Cybersecurity\n\n" \
    f"{CLI_WEBSITE_URL}\n\n"\
    f"{CLI_SUPPORT_EMAIL}\n"

# Configure options
CLI_CONFIGURE_PROXY_HOST_HELP = "Specify a proxy host for network communications to be saved into Safety's configuration. \n\n"
CLI_CONFIGURE_PROXY_PORT_HELP = "Set the proxy port to be saved into Safety's configuration file (default: 80).\n\n"
CLI_CONFIGURE_PROXY_PROTOCOL_HELP = "Choose the proxy protocol to be saved into Safety's configuration file (default: https).\n\n"
CLI_CONFIGURE_PROXY_TIMEOUT = "Set the timeout duration for proxy network calls.\n\n" + \
"[bold]Example: safety configure --proxy-timeout 30[/bold]"
CLI_CONFIGURE_PROXY_REQUIRED = "Enable or disable the requirement for a proxy in network communications\n\n" + \
"[bold]Example: safety configure --proxy-required[/bold]"
CLI_CONFIGURE_ORGANIZATION_ID = "Set the current device with an organization ID." \
" - see your Safety Platform Organization page\n\n" + \
"[bold]Example: safety configure --organization-id your_org_unique_id[/bold]"
CLI_CONFIGURE_ORGANIZATION_NAME = "Set the current device with an organization name." \
" - see your Safety Platform Organization page.\n\n" + \
"[bold]Example: safety configure --organization-name \"Your Org Name\"[/bold]"
CLI_CONFIGURE_SAVE_TO_SYSTEM = "Save the configuration to a system config file.\n" \
"This will configure Safety CLI for all users on this machine. Use --save-to-user to " \
"configure Safety CLI for only your user.\n\n" \
"[bold]Example: safety configure --save-to-system[/bold]"

# Generate options
CLI_GENERATE_PATH = "The path where the generated file will be saved (default: current directory).\n\n" \
"[bold]Example: safety generate policy_file --path .my-project-safety-policy.yml[/bold]"
CLI_GENERATE_MINIMUM_CVSS_SEVERITY = "The minimum CVSS severity to generate the installation policy for.\n\n" \
"[bold]Example: safety generate installation_policy --minimum-cvss-severity high[/bold]"

# Command default settings
CMD_PROJECT_NAME = "scan"
CMD_SYSTEM_NAME = "system-scan"
DEFAULT_CMD = CMD_PROJECT_NAME
DEFAULT_SPINNER = "bouncingBar"



import unittest
from unittest.mock import patch

from safety.tool.interceptors.types import InterceptorType
from safety.tool.interceptors.unix import UnixAliasInterceptor
from safety.tool.interceptors.windows import WindowsInterceptor
from safety.tool.interceptors.factory import create_interceptor


class TestFactory(unittest.TestCase):
    def test_explicit_unix_alias_interceptor(self):
        interceptor = create_interceptor(InterceptorType.UNIX_ALIAS)
        self.assertIsInstance(interceptor, UnixAliasInterceptor)
        
    def test_explicit_windows_interceptor(self):
        interceptor = create_interceptor(InterceptorType.WINDOWS_BAT)
        self.assertIsInstance(interceptor, WindowsInterceptor)

    @patch('safety.tool.interceptors.factory.platform', 'win32')
    def test_auto_select_windows(self):
        interceptor = create_interceptor()
        self.assertIsInstance(interceptor, WindowsInterceptor)

    def test_auto_select_unix_like(self):
        unix_platforms = ['linux', 'linux2', 'darwin']

        for platform in unix_platforms:
            with self.subTest(platform=platform):
                with patch('safety.tool.interceptors.factory.platform', 
                           platform):
                    interceptor = create_interceptor()
                    self.assertIsInstance(interceptor, UnixAliasInterceptor)                

    @patch('safety.tool.interceptors.factory.platform', 'unsupported_os')
    def test_unsupported_platform(self):
        with self.assertRaises(NotImplementedError) as context:
            create_interceptor()
        self.assertIn("Platform 'unsupported_os' is not supported", 
                     str(context.exception))

    def test_invalid_interceptor_type(self):
        invalid_type = "INVALID_TYPE"
        with self.assertRaises(KeyError):
            create_interceptor(invalid_type)



import unittest
import tempfile

from click.testing import CliRunner
from safety.auth.models import Auth
from safety.cli import cli
from safety.console import main_console as console
from unittest.mock import patch

class TestScanCommand(unittest.TestCase):

    def setUp(self):
        self.runner = CliRunner(mix_stderr=False)
        self.target = tempfile.mkdtemp()
        # Make sure the console is not quiet
        # TODO: This is a workaround, we should improve the way the console
        # is initialized in the CLI
        console.quiet = False

        cli.commands = cli.all_commands
        self.cli = cli

    @patch.object(Auth, 'is_valid', return_value=False)
    @patch('safety.auth.utils.SafetyAuthSession.get_authentication_type', return_value="unauthenticated")
    def test_scan(self, mock_is_valid, mock_get_auth_type):
        result = self.runner.invoke(self.cli, ["scan", "--target", self.target, "--output", "json"])
        self.assertEqual(result.exit_code, 1)

        result = self.runner.invoke(self.cli, ["--stage", "production", "scan", "--target", self.target, "--output", "json"])
        self.assertEqual(result.exit_code, 1)

        result = self.runner.invoke(self.cli, ["--stage", "cicd", "scan", "--target", self.target, "--output", "screen"])
        self.assertEqual(result.exit_code, 1)



from .aliases import (
    CloseResourcesEvent,
    CommandErrorEvent,
    CommandExecutedEvent,
    FirewallConfiguredEvent,
    FirewallDisabledEvent,
    FirewallHeartbeatEvent,
    FlushSecurityTracesEvent,
    PackageInstalledEvent,
    PackageUninstalledEvent,
    EventBusReadyEvent,
)
from .base import InternalEventType, InternalPayload

__all__ = [
    "CloseResourcesEvent",
    "FlushSecurityTracesEvent",
    "InternalEventType",
    "InternalPayload",
    "CommandExecutedEvent",
    "CommandErrorEvent",
    "PackageInstalledEvent",
    "PackageUninstalledEvent",
    "FirewallHeartbeatEvent",
    "FirewallConfiguredEvent",
    "FirewallDisabledEvent",
    "EventBusReadyEvent",
]



import os
import unittest
from unittest.mock import MagicMock, Mock, call, patch
from pathlib import Path
import datetime

from safety.scan.render import print_announcements, print_summary, render_header, prompt_project_id
from safety_schemas.models import ProjectModel, IgnoreCodes, PolicySource

class TestRender(unittest.TestCase):

    def setUp(self):
        self.console = MagicMock()
        self.project = ProjectModel(id='test-project')
        self.project.policy = MagicMock()
        self.project.policy.source = PolicySource.cloud

    @patch('safety.scan.render.get_version')
    def test_render_header(self, mock_get_safety_version):
        mock_get_safety_version.return_value = '3.0.0'

        datetime_mock = Mock(wraps=datetime.datetime)
        datetime_mock.now.return_value = datetime.datetime(2025, 1, 1, 0, 0, 0, tzinfo=datetime.timezone.utc)

        test_cases = [
            ([Path('/target1'), Path('/target2')], False, 'Safety 3.0.0 scanning {0}, {1}\n2025-01-01 00:00:00 UTC'),
            ([Path('/target1')], False, 'Safety 3.0.0 scanning {0}\n2025-01-01 00:00:00 UTC'),
            ([], False, 'Safety 3.0.0 scanning \n2025-01-01 00:00:00 UTC'),
            ([Path('/target1'), Path('/target2')], True, 'Safety 3.0.0 running system scan\n2025-01-01 00:00:00 UTC'),
        ]

        for targets, is_system_scan, expected_result in test_cases:
            with patch('datetime.datetime', new=datetime_mock):
                # Normalize paths to POSIX format (use forward slashes)
                posix_targets = [t.as_posix() for t in targets]
                result = str(render_header(targets, is_system_scan))

                # On Windows, convert backslashes to forward slashes in the result for comparison
                if os.name == 'nt':
                    result = result.replace('\\', '/')

                expected = expected_result.format(*posix_targets)
                self.assertEqual(result, expected)

    @patch('safety.scan.render.safety.get_announcements')
    @patch('safety.scan.render.get_basic_announcements')
    @patch('safety.scan.render.Console')
    def test_print_announcements(self, mock_console, mock_get_basic_announcements, mock_get_announcements):
        
        mock_get_announcements.return_value = [
            {'type': 'info', 'message': 'Info message'},
            {'type': 'warning', 'message': 'Warning message'},
            {'type': 'error', 'message': 'Error message'},
        ]
        
        mock_get_basic_announcements.return_value = mock_get_announcements.return_value
        console = mock_console.return_value
        
        ctx = MagicMock()
        ctx.obj.auth.client = MagicMock()
        ctx.obj.config.telemetry_enabled = False
        ctx.obj.telemetry = MagicMock()
        
        print_announcements(console, ctx)

        # Include empty calls and correct sequence
        console.print.assert_has_calls([
            call(),
            call("[bold]Safety Announcements:[/bold]"),
            call(),
            call("[default]* Info message[/default]"),
            call("[yellow]* Warning message[/yellow]"),
            call("[red]* Error message[/red]")
        ])


    @patch('safety.scan.render.render_to_console')
    def test_print_summary(self, mock_render_to_console):
        ignored_vulns_data = [
            MagicMock(ignored_code=IgnoreCodes.manual.value, vulnerability_id='v1', package_name='p1'),
            MagicMock(ignored_code=IgnoreCodes.cvss_severity.value, vulnerability_id='v2', package_name='p2'),
            MagicMock(ignored_code=IgnoreCodes.unpinned_specification.value, vulnerability_id='v3', package_name='p3'),
            MagicMock(ignored_code=IgnoreCodes.environment_dependency.value, vulnerability_id='v4', package_name='p4'),
        ]

        print_summary(
            self.console,
            total_issues_with_duplicates=0,
            total_ignored_issues=0,
            project=self.project,
            dependencies_count=5,
            fixes_count=0,
            resolved_vulns_per_fix=0,
            ignored_vulns_data=ignored_vulns_data
        )

        self.console.print.assert_has_calls([
            call('Tested [number]5[/number] dependencies for security issues using policy fetched from Safety Platform'),
            call('0 security issues found, 0 fixes suggested.'),
            call('[number]0[/number] fixes suggested, resolving [number]0[/number] vulnerabilities.')
        ])

        print_summary(
            self.console,
            total_issues_with_duplicates=0,
            total_ignored_issues=0,
            project=self.project,
            dependencies_count=5,
            fixes_count=0,
            resolved_vulns_per_fix=0
        )

        self.console.print.assert_has_calls([
            call('Tested [number]5[/number] dependencies for security issues using policy fetched from Safety Platform'),
            call('0 security issues found, 0 fixes suggested.'),
            call('[number]0[/number] fixes suggested, resolving [number]0[/number] vulnerabilities.')
        ])

    @patch("safety.scan.render.clean_project_id")
    def test_prompt_project_id_non_interactive(self, clean_project_id):
        """
        Under these cases, the default project ID should be cleaned and 
        returned. The prompt should not be shown.
        """
        
        test_cases = [
            # Non-interactive mode
            (True, False, "default_a", "default_a_cleaned"),
            # Quiet mode like JSON output under interactive mode
            (True, True, "default_b", "default_b_cleaned"),
            # No Quiet and Not interactive mode
            (False, False, "default_c", "default_c_cleaned"),
        ]
        
        for quiet, is_interactive, default_id, expected_result in test_cases:
            with self.subTest(quiet=quiet, is_interactive=is_interactive):        
        
                clean_project_id.return_value = f"{default_id}_cleaned"    
                console = MagicMock(quiet=quiet, is_interactive=is_interactive)
                
                result = prompt_project_id(console, default_id)

                assert result == expected_result

                assert result == expected_result, (
                    f"Failed for quiet={quiet}, "
                    f"is_interactive={is_interactive}\n"
                    f"Expected: {expected_result}\n"
                    f"Got: {result}\n"
                    f"Default ID was: {default_id}"
                )

                try:
                    clean_project_id.assert_called_once_with(default_id)
                except AssertionError:
                    raise AssertionError(
                        f"Mock wasn't called correctly for "
                        f"quiet={quiet}, is_interactive={is_interactive}\n"
                        f"Expected (1) call with: {default_id}\n"
                        f"Actual calls were "
                        f"({len(clean_project_id.call_args_list)}): "
                        f"{clean_project_id.call_args_list}"
                    )

                clean_project_id.reset_mock()


    @patch("safety.scan.render.clean_project_id")
    def test_prompt_project_id_interactive(self, clean_project_id):
        default_id = "default-project"
        default_id_cleaned = f"{default_id}_cleaned"
        
        test_cases = [("custom-project", "custom-project_cleaned"),
                      ("", default_id_cleaned)]

        for user_input, expected in test_cases:
            with self.subTest(user_input=user_input):
                console = MagicMock(quiet=False, is_interactive=True)

                clean_project_id.side_effect = lambda input_string: (
                    f"{input_string}_cleaned"
                )
                
                with patch('safety.scan.render.Prompt.ask') as ask:

                    # We mimic the behavior of Prompt.ask on empty input
                    ask.side_effect = lambda *args, **kwargs: (
                        kwargs['default'] if user_input == "" else user_input
                    )
                    
                    result = prompt_project_id(console, default_id)

                    # Verify Prompt.ask was called correctly
                    ask.assert_called_once_with(
                        f"Set a project id (no spaces). If empty Safety will use [bold]{default_id_cleaned}[/bold]",
                        console=console,
                        default=default_id_cleaned,
                        show_default=False
                    )

                    calls = [call(default_id)]
                    call_count = 1

                    if user_input != "":
                        calls.append(call(user_input))
                        call_count = 2
                    
                    clean_project_id.assert_has_calls(calls)                    
                    assert clean_project_id.call_count == call_count

                    print(result, expected)                    
                    assert result == expected
                    
                # Reset mocks for next test case
                clean_project_id.reset_mock()



# type: ignore
import os
import sys
import unittest
from io import StringIO
from unittest.mock import MagicMock, patch, Mock

import click as click

from safety import util
from safety.models import SafetyRequirement
from safety.util import (
    read_requirements,
    get_processed_options,
    SafetyPolicyFile,
    transform_ignore,
)


class ReadRequirementsTestCase(unittest.TestCase):
    def setUp(self) -> None:
        self.dirname = os.path.dirname(__file__)

    def test_unpinned_vcs_requirement(self):
        """
        https://github.com/pyupio/safety/issues/72
        """
        # this shouldn't raise an error
        content = StringIO(
            "-e git+https://github.com/jdunck/python-unicodecsv#egg=unicodecsv"
        )
        result = list(read_requirements(content))
        self.assertEqual(len(result), 0)

    def test_recursive_requirement(self):
        """
        https://github.com/pyupio/safety/issues/132
        """
        # this should find 1 packages with two requirements found
        dirname = os.path.dirname(__file__)
        test_filename = os.path.join(dirname, "reqs_1.txt")
        with open(test_filename) as fh:
            result = list(read_requirements(fh, resolve=True))
        self.assertEqual(len(result), 1)
        found_pkg = result[0]
        self.assertEqual(found_pkg.name, "insecure-package")
        self.assertEqual(found_pkg.version, None)
        expected = [
            SafetyRequirement("insecure-package==0.1.0"),
            SafetyRequirement("insecure-package==0.1.1"),
        ]
        self.assertListEqual(expected, found_pkg.requirements)

    def test_recursive_requirement_pinned_after_unpinned(self):
        # this should find 4 packages, unpinned aren't ignored.
        dirname = os.path.dirname(__file__)
        test_filename = os.path.join(dirname, "reqs_pinned_and_unpinned.txt")
        with open(test_filename) as fh:
            result = list(read_requirements(fh, resolve=True))
        self.assertEqual(len(result), 4)

    @patch("safety.util.get_flags_from_context")
    @patch.object(
        sys,
        "argv",
        [
            "safety/__main__.py",
            "check",
            "--key=my-key",
            "-i",
            "3232",
            "-i",
            "3231",
            "--ignore",
            "1212",
            "--json",
        ],
    )
    def test_log_used_options_with_argv(self, get_flags_from_context):
        get_flags_from_context.return_value = {
            "--key": "key",
            "--db": "db",
            "--json": "json",
            "--no-json": "json",
            "--full-report": "full_report",
            "--short-report": "full_report",
            "--bare": "bare",
            "--not-bare": "bare",
            "--cache": "cache",
            "--no-cache": "cache",
            "--stdin": "stdin",
            "--no-stdin": "stdin",
            "--file": "files",
            "-r": "files",
            "--ignore": "ignore",
            "-i": "ignore",
            "--output": "output",
            "-o": "output",
            "--proxy-host": "proxyhost",
            "-ph": "proxyhost",
            "--proxy-port": "proxyport",
            "-pp": "proxyport",
            "--proxy-protocol": "proxyprotocol",
            "-pr": "proxyprotocol",
        }
        used_options = util.get_used_options()

        self.assertEqual(
            used_options,
            {
                "ignore": {"-i": 2, "--ignore": 1},
                "json": {"--json": 1},
                "key": {"--key": 1},
            },
        )

    @patch.object(
        click,
        "get_current_context",
        Mock(
            get_parameter_source=Mock(return_value=click.core.ParameterSource.DEFAULT)
        ),
    )
    def test_cli_ignore_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_with_ignores.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        cli_ignores = {"1234": {"reason": "", "expires": None}}
        ignore, ignore_severity_rules, exit_code, ignore_unpinned_packages, project = (
            get_processed_options(
                policy_file=policy_file,
                ignore=cli_ignores,
                ignore_severity_rules=None,
                exit_code=True,
            )
        )

        self.assertEqual(ignore, cli_ignores)

    @patch.object(
        click,
        "get_current_context",
        Mock(
            get_parameter_source=Mock(
                return_value=click.core.ParameterSource.COMMANDLINE
            )
        ),
    )
    def test_cli_continue_on_error_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_full.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        ignore, ignore_severity_rules, exit_code, ignore_unpinned_packages, project = (
            get_processed_options(
                policy_file=policy_file,
                ignore={},
                ignore_severity_rules=None,
                exit_code=True,
            )
        )

        self.assertEqual(exit_code, True)

    @patch.object(
        click,
        "get_current_context",
        Mock(
            get_parameter_source=Mock(
                return_value=click.core.ParameterSource.COMMANDLINE
            )
        ),
    )
    def test_cli_exit_code_partial_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_full.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        # Cli only passes the exit_code argument by commandline
        ignore, ignore_severity_rules, exit_code, ignore_unpinned_packages, project = (
            get_processed_options(
                policy_file=policy_file,
                ignore={},
                ignore_severity_rules=None,
                exit_code=True,
            )
        )

        security_pf = policy_file.get("security", {})
        severity_rules = {
            "ignore-cvss-severity-below": security_pf.get(
                "ignore-cvss-severity-below", 0.0
            ),
            "ignore-cvss-unknown-severity": security_pf.get(
                "ignore-cvss-unknown-severity", False
            ),
        }

        self.assertEqual(ignore, security_pf.get("ignore-vulnerabilities", None))
        self.assertEqual(ignore_severity_rules, severity_rules)
        self.assertEqual(exit_code, True)

    @patch.object(
        click,
        "get_current_context",
        Mock(
            get_parameter_source=Mock(return_value=click.core.ParameterSource.DEFAULT)
        ),
    )
    def test_cli_ignore_partial_overrule_policy_file(self):
        path_pf = os.path.join(self.dirname, ".policy_full.yml")
        policy_file = SafetyPolicyFile().convert(value=path_pf, param=None, ctx=None)

        # Cli only passes the ignores argument by commandline
        cli_ignores = {"1234": {"reason": "", "expires": None}}
        ignore, ignore_severity_rules, exit_code, ignore_unpinned_packages, project = (
            get_processed_options(
                policy_file=policy_file,
                ignore=cli_ignores,
                ignore_severity_rules=None,
                exit_code=True,
            )
        )

        security_pf = policy_file.get("security", {})
        severity_rules = {
            "ignore-cvss-severity-below": security_pf.get(
                "ignore-cvss-severity-below", 0.0
            ),
            "ignore-cvss-unknown-severity": security_pf.get(
                "ignore-cvss-unknown-severity", False
            ),
        }

        self.assertEqual(ignore, cli_ignores)
        self.assertEqual(ignore_severity_rules, severity_rules)
        self.assertIsNotNone(
            security_pf.get("continue-on-vulnerability-error", None),
            msg="This test requires a yml with a continue-on-vulnerability-error value",
        )
        EXPECTED = not security_pf.get("continue-on-vulnerability-error")
        self.assertEqual(exit_code, EXPECTED)

    def test_transform_ignore(self):
        ignored_transformed = {
            "123": {"expires": None, "reason": ""},
            "456": {"expires": None, "reason": ""},
        }
        self.assertEqual(
            transform_ignore(None, None, value=("123", "456")), ignored_transformed
        )
        self.assertEqual(
            transform_ignore(None, None, value=("123,456",)), ignored_transformed
        )

    def test_transform_ignore_mixed_arguments(self):
        # mix old and new way of providing --ignore
        ignored_transformed = {
            "123": {"expires": None, "reason": ""},
            "456": {"expires": None, "reason": ""},
            "789": {"expires": None, "reason": ""},
        }
        self.assertEqual(
            transform_ignore(None, None, value=("123,456", "789")), ignored_transformed
        )


class TestInitializeEventBus(unittest.TestCase):
    def setUp(self):
        self.mock_ctx = MagicMock()
        self.mock_obj = MagicMock()
        self.mock_auth = MagicMock()
        self.mock_client = MagicMock()

        self.mock_ctx.obj = self.mock_obj
        self.mock_obj.auth = self.mock_auth
        self.mock_auth.client = self.mock_client
        self.mock_obj.events_enabled = True
        self.mock_obj.event_bus = MagicMock()

    @patch("safety.util.start_event_bus")
    @patch("safety.events.utils.create_internal_event")
    def test_successful_initialization_with_token(
        self, mock_create_event, mock_start_event_bus
    ):
        self.mock_client.token = {"access_token": "test_token"}
        self.mock_client.api_key = None
        mock_event = MagicMock()
        mock_create_event.return_value = mock_event

        from safety.util import initialize_event_bus

        result = initialize_event_bus(self.mock_ctx)

        self.assertTrue(result)
        mock_start_event_bus.assert_called_once_with(self.mock_obj, self.mock_client)
        self.mock_obj.event_bus.emit.assert_called_once_with(mock_event)

    @patch("safety.util.start_event_bus")
    @patch("safety.events.utils.create_internal_event")
    def test_successful_initialization_with_api_key(
        self, mock_create_event, mock_start_event_bus
    ):
        self.mock_client.token = None
        self.mock_client.api_key = "test_api_key"
        mock_event = MagicMock()
        mock_create_event.return_value = mock_event

        from safety.util import initialize_event_bus

        result = initialize_event_bus(self.mock_ctx)

        self.assertTrue(result)
        mock_start_event_bus.assert_called_once_with(self.mock_obj, self.mock_client)
        self.mock_obj.event_bus.emit.assert_called_once_with(mock_event)

    @patch("safety.util.start_event_bus")
    def test_initialization_when_no_authn(self, mock_start_event_bus):
        self.mock_client.token = None
        self.mock_client.api_key = None

        from safety.util import initialize_event_bus

        result = initialize_event_bus(self.mock_ctx)

        self.assertFalse(result)
        mock_start_event_bus.assert_not_called()

    @patch("safety.util.LOG")
    def test_initialization_with_exception(self, mock_log):
        self.mock_client.token = {"access_token": "test_token"}
        self.mock_obj.event_bus = None

        with patch(
            "safety.util.start_event_bus", side_effect=Exception("Test exception")
        ):
            from safety.util import initialize_event_bus

            result = initialize_event_bus(self.mock_ctx)

        self.assertFalse(result)
        mock_log.exception.assert_called_once()



import json
import logging
import os
import shutil
import socket
import sys
import tempfile
import unittest
from pathlib import Path
from unittest.mock import Mock, patch

import click
import psutil
import requests
from click.testing import CliRunner
from packaging.version import Version
from safety_schemas.models.base import AuthenticationType

from safety.auth.models import Auth
from safety.cli import cli
from safety.console import main_console as console
from safety.meta import get_version
from safety.models import CVE, SafetyCLI, SafetyRequirement, Severity, Vulnerability
from safety.util import Package, SafetyContext


def get_vulnerability(vuln_kwargs=None, cve_kwargs=None, pkg_kwargs=None):
    vuln_kwargs = {} if vuln_kwargs is None else vuln_kwargs
    cve_kwargs = {} if cve_kwargs is None else cve_kwargs
    pkg_kwargs = {} if pkg_kwargs is None else pkg_kwargs

    p_kwargs = {'name': 'django', 'version': '2.2', 'requirements': [SafetyRequirement('django==2.2')], 'found': '/site-packages/django',
                'insecure_versions': [], 'secure_versions': ['2.2'],
                'latest_version_without_known_vulnerabilities': '2.2',
                'latest_version': '2.2', 'more_info_url': 'https://pyup.io/package/foo'}
    p_kwargs.update(pkg_kwargs)

    c_kwargs = {'name': 'bla', 'cvssv2': None, 'cvssv3': None}
    c_kwargs.update(cve_kwargs)

    cve = CVE(**c_kwargs)
    severity = None
    if cve and cve.cvssv2 or cve.cvssv3:
        severity = Severity(source=cve.name, cvssv2=cve.cvssv2, cvssv3=cve.cvssv3)
    pkg = Package(**p_kwargs)

    vulnerable_spec = set()
    vulnerable_spec.add(">0")

    v_kwargs = {'package_name': pkg.name, 'pkg': pkg, 'ignored': False, 'ignored_reason': '', 'ignored_expires': '',
                'vulnerable_spec': vulnerable_spec,
                'all_vulnerable_specs': ['2.2'],
                'analyzed_version': pkg.version,
                'analyzed_requirement': pkg.requirements[0],
                'advisory': '',
                'vulnerability_id': 'PYUP-1234',
                'is_transitive': False,
                'published_date': '',
                'fixed_versions': [],
                'closest_versions_without_known_vulnerabilities': '',
                'resources': ["pyup.io/vuln-id"],
                'CVE': cve,
                'severity': severity,
                'affected_versions': [],
                'more_info_url': 'https://pyup.io/PVE/2323'}

    v_kwargs.update(vuln_kwargs)

    return Vulnerability(**v_kwargs)


class TestSafetyCLI(unittest.TestCase):

    def setUp(self):
        self.maxDiff = None
        self.runner = CliRunner(mix_stderr=False)
        self.output_options = ['screen', 'text', 'json', 'bare']
        self.dirname = os.path.dirname(__file__)
        # Make sure the console is not quiet
        # TODO: This is a workaround, we should improve the way the console
        # is initialized in the CLI
        console.quiet = False

        # Reset the commands
        # CLI initialization is made on the import of the module
        # so we need to reset the commands to avoid side effects
        # 
        # TODO: this is a workaround, we should improve the way the 
        # CLI is initialized
        cli.commands = cli.all_commands
        self.cli = cli


    def test_command_line_interface(self):
        runner = CliRunner()
        result = runner.invoke(self.cli)
        expected = "Usage: cli [OPTIONS] COMMAND [ARGS]..."

        for option in [[], ["--help"]]:
            result = runner.invoke(self.cli, option)
            self.assertEqual(result.exit_code, 0)
            self.assertIn(expected, click.unstyle(result.output))

    @patch("safety.safety.check")
    def test_check_vulnerabilities_found_default(self, check_func):
        check_func.return_value = [get_vulnerability()], None
        EXPECTED_EXIT_CODE_VULNS_FOUND = 64
        result = self.runner.invoke(self.cli, ['check'])
        self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_VULNS_FOUND)

    @patch("safety.safety.check")
    def test_check_vulnerabilities_not_found_default(self, check_func):
        check_func.return_value = [], None
        EXPECTED_EXIT_CODE_VULNS_NOT_FOUND = 0
        result = self.runner.invoke(self.cli, ['check'])
        self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_VULNS_NOT_FOUND)

    @patch("safety.safety.check")
    def test_check_vulnerabilities_found_with_outputs(self, check_func):
        check_func.return_value = [get_vulnerability()], None
        EXPECTED_EXIT_CODE_VULNS_FOUND = 64

        for output in self.output_options:
            result = self.runner.invoke(self.cli, ['check', '--output', output])
            self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_VULNS_FOUND)

    @patch("safety.safety.check")
    def test_check_vulnerabilities_not_found_with_outputs(self, check_func):
        check_func.return_value = [], None
        EXPECTED_EXIT_CODE_VULNS_NOT_FOUND = 0

        for output in self.output_options:
            result = self.runner.invoke(self.cli, ['check', '--output', output])
            self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_VULNS_NOT_FOUND)

    @patch("safety.safety.check")
    def test_check_continue_on_error(self, check_func):
        EXPECTED_EXIT_CODE_CONTINUE_ON_ERROR = 0

        # Cover no vulns found and vulns found
        for vulns in [[get_vulnerability()], []]:
            check_func.return_value = vulns, None

            result = self.runner.invoke(self.cli, ['check', '--continue-on-error'])
            self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_CONTINUE_ON_ERROR)

            for output in self.output_options:
                result = self.runner.invoke(self.cli, ['check', '--output', output, '--continue-on-error'])
                self.assertEqual(result.exit_code, EXPECTED_EXIT_CODE_CONTINUE_ON_ERROR)

    @patch("safety.safety.get_announcements")
    def test_announcements_if_is_not_tty(self, get_announcements_func):
        announcement = {'type': 'error', 'message': 'Please upgrade now'}
        get_announcements_func.return_value = [announcement]
        message = f"* {announcement.get('message')}"
        result = self.runner.invoke(self.cli, ['check'])
        self.assertTrue('ANNOUNCEMENTS' in result.stderr)
        self.assertTrue(message in result.stderr)


    @patch("safety.safety.check")
    def test_check_ignore_format_backward_compatible(self, check):
        runner = CliRunner()

        check.return_value = []

        dirname = os.path.dirname(__file__)
        reqs_path = os.path.join(dirname, "reqs_4.txt")

        _ = runner.invoke(self.cli, ['check', '--file', reqs_path, '--ignore', "123,456", '--ignore', "789"])
        try:
            check_call_kwargs = check.call_args[1]  # Python < 3.8
        except IndexError:
            check_call_kwargs = check.call_args.kwargs

        ignored_transformed = {
            '123': {'expires': None, 'reason': ''},
            '456': {'expires': None, 'reason': ''},
            '789': {'expires': None, 'reason': ''}
        }
        self.assertEqual(check_call_kwargs['ignore_vulns'], ignored_transformed)

    def test_validate_with_unsupported_argument(self):
        result = self.runner.invoke(self.cli, ['validate', 'safety_ci'])
        msg = 'This Safety version only supports "policy_file" validation. "safety_ci" is not supported.\n'
        self.assertEqual(click.unstyle(result.stderr), msg)
        self.assertEqual(result.exit_code, 1)

    def test_validate_with_wrong_path(self):
        p = Path('imaginary/path')
        result = self.runner.invoke(self.cli, ['validate', 'policy_file', '--path', str(p)])
        msg = f'The path "{str(p)}" does not exist.\n'
        self.assertEqual(click.unstyle(result.stderr), msg)
        self.assertEqual(result.exit_code, 1)

    def test_validate_with_basic_policy_file(self):
        dirname = os.path.dirname(__file__)

        # Test with policy version 2.0
        path = os.path.join(dirname, "test_policy_file", "default_policy_file.yml")
        result = self.runner.invoke(self.cli, ['validate', 'policy_file', '2.0', '--path', path])
        cleaned_stdout = click.unstyle(result.stdout)
        msg = 'The Safety policy file (Valid only for the check command) was successfully parsed with the following values:\n'
        parsed = json.dumps(
            {
                "project-id": '',
                "security": {
                    "ignore-cvss-severity-below": 0,
                    "ignore-cvss-unknown-severity": False,
                    "ignore-vulnerabilities": {
                        "25853": {
                            "reason": "we don't use the vulnerable function",
                            "expires": "2022-10-21 00:00:00"
                        }
                    },
                    "continue-on-vulnerability-error": False
                },
                "filename": path
            },
            indent=2
        ) + '\n'

        self.assertEqual(msg + parsed, cleaned_stdout)
        self.assertEqual(result.exit_code, 0)

        # Test with policy version 3.0
        path = os.path.join(dirname, "test_policy_file", "v3_0", "default_policy_file.yml")
        result = self.runner.invoke(self.cli, ['validate', 'policy_file', '3.0', '--path', path])
        cleaned_stdout = click.unstyle(result.stdout)
        msg = 'The Safety policy (3.0) file (Used for scan and system-scan commands) was successfully parsed with the following values:\n'

        parsed = {
            "version": "3.0",
            "scan": {
                "max_depth": 6,
                "exclude": [],
                "include_files": [],
                "system": {
                    "targets": []
                }
            },
            "report": {
                "dependency_vulnerabilities": {
                    "enabled": True,
                    "auto_ignore": {
                        "python": {
                            "ignore_environment_results": True,
                            "ignore_unpinned_requirements": True
                        },
                        "vulnerabilities": None,
                        "cvss_severity": []
                    }
                }
            },
            "fail_scan": {
                "dependency_vulnerabilities": {
                    "enabled": True,
                    "fail_on_any_of": {
                        "cvss_severity": [
                            "critical",
                            "high",
                            "medium",
                        ],
                        "exploitability": [
                            "critical",
                            "high",
                            "medium",
                        ]
                    }
                }
            },
            "security_updates": {
                "dependency_vulnerabilities": {
                    "auto_security_updates_limit": [
                        "patch"
                    ]
                }
            },
        }

        msg_stdout, parsed_policy = cleaned_stdout.split('\n', 1)
        msg_stdout += '\n'
        parsed_policy = json.loads(parsed_policy.replace('\n', ''))

        # Remove the 'installation' key if it exists
        parsed_policy.pop("installation", None)

        # Sorting and comparing specific fields
        fail_scan = parsed_policy.get("fail_scan", None)
        self.assertIsNotNone(fail_scan)
        fail_of_any = fail_scan["dependency_vulnerabilities"]["fail_on_any_of"]
        fail_of_any["cvss_severity"] = sorted(fail_of_any["cvss_severity"])
        fail_of_any["exploitability"] = sorted(fail_of_any["exploitability"])

        # Assert that the message is the same
        self.assertEqual(msg, msg_stdout)

        # Assert that the parsed policy matches the expected policy
        self.assertEqual(parsed, parsed_policy)

        # Check the exit code
        self.assertEqual(result.exit_code, 0)


    def test_validate_with_policy_file_using_invalid_keyword(self):
        dirname = os.path.dirname(__file__)
        filename = 'default_policy_file_using_invalid_keyword.yml'
        path = os.path.join(dirname, "test_policy_file", filename)
        result = self.runner.invoke(self.cli, ['validate', 'policy_file', '2.0', '--path', path])
        cleaned_stdout = click.unstyle(result.stderr)
        msg_hint = 'HINT: "security" -> "transitive" is not a valid keyword. Valid keywords in this level are: ' \
                   'ignore-cvss-severity-below, ignore-cvss-unknown-severity, ignore-vulnerabilities, ' \
                   'continue-on-vulnerability-error, ignore-unpinned-requirements\n'
        msg = f'Unable to load the Safety Policy file "{path}".\n{msg_hint}'

        self.assertEqual(msg, cleaned_stdout)
        self.assertEqual(result.exit_code, 1)

        path = os.path.join(dirname, "test_policy_file", "v3_0", filename)
        result = self.runner.invoke(self.cli, ['validate', 'policy_file', '3.0', '--path', path])
        cleaned_stdout = click.unstyle(result.stderr)
        msg = f'Unable to load the Safety Policy file ("{path}"), this command only supports version 3.0, details: 1 validation error for Config'

        self.assertIn(msg, cleaned_stdout)
        self.assertEqual(result.exit_code, 1)


    def test_validate_with_policy_file_using_invalid_typo_keyword(self):
        dirname = os.path.dirname(__file__)
        filename = 'default_policy_file_using_invalid_typo_keyword.yml'
        path = os.path.join(dirname, "test_policy_file", filename)
        result = self.runner.invoke(self.cli, ['validate', 'policy_file', '2.0', '--path', path])
        cleaned_stdout = click.unstyle(result.stderr)
        msg_hint = 'HINT: "security" -> "ignore-vunerabilities" is not a valid keyword. Maybe you meant: ' \
                   'ignore-vulnerabilities\n'
        msg = f'Unable to load the Safety Policy file "{path}".\n{msg_hint}'

        self.assertEqual(msg, cleaned_stdout)
        self.assertEqual(result.exit_code, 1)

    def test_generate_pass(self):
        with tempfile.TemporaryDirectory() as tempdir:
            result = self.runner.invoke(self.cli, ['generate', 'policy_file', '--path', tempdir])
            cleaned_stdout = click.unstyle(result.stdout)
            msg = f'A default Safety policy file has been generated! Review the file contents in the path {tempdir} ' \
                  f'in the file: .safety-policy.yml\n'
            self.assertEqual(msg, cleaned_stdout)

    def test_generate_with_unsupported_argument(self):
        result = self.runner.invoke(self.cli, ['generate', 'safety_ci'])
        msg = 'This Safety version only supports "policy_file" generation. "safety_ci" is not supported.\n'
        self.assertEqual(click.unstyle(result.stderr), msg)
        self.assertEqual(result.exit_code, 1)

    def test_generate_with_wrong_path(self):
        p = Path('imaginary/path')
        result = self.runner.invoke(self.cli, ['generate', 'policy_file', '--path', str(p)])
        msg = f'The path "{str(p)}" does not exist.\n'
        self.assertEqual(click.unstyle(result.stderr), msg)
        self.assertEqual(result.exit_code, 1)

    def test_check_with_fix_does_verify_api_key(self):
        dirname = os.path.dirname(__file__)
        req_file = os.path.join(dirname, "test_fix", "basic", "reqs_simple.txt")
        result = self.runner.invoke(self.cli, ['check', '-r', req_file, '--apply-security-updates'])
        self.assertEqual(click.unstyle(result.stderr),
                         "The --apply-security-updates option needs authentication. See https://docs.safetycli.com/safety-docs/support/invalid-api-key-error.\n")
        self.assertEqual(result.exit_code, 65)

    def test_check_with_fix_only_works_with_files(self):
        result = self.runner.invoke(self.cli, ['check', '--key', 'TEST-API_KEY', '--apply-security-updates'])
        self.assertEqual(click.unstyle(result.stderr),
                         '--apply-security-updates only works with files; use the "-r" option to specify files to remediate.\n')
        self.assertEqual(result.exit_code, 1)

    @patch("safety.util.SafetyContext")
    @patch("safety.safety.check")
    @patch("safety.safety.calculate_remediations")
    @patch("safety.cli.get_packages")
    def test_check_with_fix(self, get_packages, calculate_remediations, check_func, ctx):
        vulns = [get_vulnerability()]
        packages = [pkg for pkg in {vuln.pkg.name: vuln.pkg for vuln in vulns}.values()]
        get_packages.return_value = packages
        provided_context = SafetyContext()
        provided_context.command = 'check'
        provided_context.packages = packages
        ctx.return_value = provided_context
        check_func.return_value = vulns, None
        target = Version("1.9")
        calculate_remediations.return_value = {
            "django": {
                "==1.8": {
                    "version": "1.8",
                    "vulnerabilities_found": 1,
                    "recommended_version": target,
                    "requirement": SafetyRequirement('django==1.8'),
                    "secure_versions": [],
                    "closest_secure_version": {"minor": None, "major": target},
                    "more_info_url": "https://pyup.io/p/pypi/django/52d/"}
            }}

        dirname = os.path.dirname(__file__)
        source_req = os.path.join(dirname, "test_fix", "basic", "reqs_simple.txt")

        with tempfile.TemporaryDirectory() as tempdir:
            req_file = os.path.join(tempdir, 'reqs_simple_minor.txt')
            shutil.copy(source_req, req_file)

            self.runner.invoke(self.cli, ['check', '-r', req_file, '--key', 'TEST-API_KEY',
                                             '--apply-security-updates'])

            with open(req_file) as f:
                self.assertEqual("django==1.8\nsafety==2.3.0\nflask==0.87.0", f.read())

            self.runner.invoke(self.cli, ['check', '-r', req_file, '--key', 'TEST-API_KEY', '--apply-security-updates',
                                         '--auto-security-updates-limit', 'minor'])

            with open(req_file) as f:
                self.assertEqual("django==1.9\nsafety==2.3.0\nflask==0.87.0", f.read())

            target = Version("2.0")
            calculate_remediations.return_value = {
                "django": {
                    "==1.9": {
                        "version": "1.9",
                        "vulnerabilities_found": 1,
                        "recommended_version": target,
                        "requirement": SafetyRequirement('django==1.9'),
                        "secure_versions": [],
                        "closest_secure_version": {"minor": None, "major": target},
                        "more_info_url": "https://pyup.io/p/pypi/django/52d/"}}
            }

            self.runner.invoke(self.cli, ['check', '-r', req_file, '--key', 'TEST-API_KEY', '--apply-security-updates',
                                         '-asul', 'minor', '--json'])
            with open(req_file) as f:
                self.assertEqual("django==1.9\nsafety==2.3.0\nflask==0.87.0", f.read())

            self.runner.invoke(self.cli, ['check', '-r', req_file, '--key', 'TEST-API_KEY', '--apply-security-updates',
                                         '-asul', 'major', '--output', 'bare'])

            with open(req_file) as f:
                self.assertEqual("django==2.0\nsafety==2.3.0\nflask==0.87.0", f.read())

    def test_check_ignore_unpinned_requirements(self):
        dirname = os.path.dirname(__file__)
        db = os.path.join(dirname, "test_db")
        reqs_unpinned = os.path.join(dirname, "reqs_unpinned.txt")

        # Test default behavior (ignore_unpinned_requirements is None)
        result = self.runner.invoke(self.cli, ['check', '-r', reqs_unpinned, '--db', db, '--output', 'text'])

        # Check for deprecation message
        self.assertIn("DEPRECATED: this command (`check`) has been DEPRECATED", result.output)

        # Check for the announcement about unpinned packages
        expected_announcement = "Warning: django and numpy are unpinned. Safety by default does not report"
        self.assertIn(expected_announcement, result.output)

        # Check for the warning about potential vulnerabilities
        expected_warning = "Warning: 2 known vulnerabilities match the django versions that could be"
        self.assertIn(expected_warning, result.output)

        # Test ignore_unpinned_requirements set to True
        result = self.runner.invoke(self.cli, ['check', '-r', reqs_unpinned, '--ignore-unpinned-requirements',
                                            '--db', db, '--output', 'text'])

        self.assertIn("Warning: django and numpy are unpinned and potential vulnerabilities are", result.output)
        self.assertIn("being ignored given `ignore-unpinned-requirements` is True in your config.", result.output)

        # Test check_unpinned_requirements set to True
        result = self.runner.invoke(self.cli, ['check', '-r', reqs_unpinned, '--db', db, '--json', '-i', 'some id',
                                          '--check-unpinned-requirements'])

        # Check for deprecation message
        self.assertIn("DEPRECATED: this command (`check`) has been DEPRECATED", result.output)

        # Extract JSON part from the output
        json_start = result.output.find('{')
        json_end = result.output.rfind('}') + 1
        json_output = result.output[json_start:json_end]

        try:
            parsed_json = json.loads(json_output)
            vulnerabilities = parsed_json.get('vulnerabilities', [])
            self.assertEqual(1, len(vulnerabilities), 'Unexpected number of vulnerabilities reported.')

            ignored = parsed_json.get('ignored_vulnerabilities', [])
            self.assertEqual(1, len(ignored), 'Unexpected number of ignored vulnerabilities.')

            reason = ignored[0].get('ignored_reason', None)
            self.assertEqual("", reason, "Unexpected ignore reason.")
        except json.JSONDecodeError:
            self.fail(f"Failed to parse JSON output. Extracted JSON was: {json_output}")

    def test_basic_html_output_pass(self):
        dirname = os.path.dirname(__file__)
        db = os.path.join(dirname, "test_db")
        reqs_unpinned = os.path.join(dirname, "reqs_unpinned.txt")

        result = self.runner.invoke(self.cli, ['check', '-r', reqs_unpinned, '--db', db, '--output', 'html'])

        ignored = "<p>Found vulnerabilities that were ignored: 2</p>"
        announcement = "Warning: django and numpy are unpinned."
        self.assertIn(ignored, result.stdout)
        self.assertIn(announcement, result.stdout)
        self.assertNotIn("remediations-suggested", result.stdout)

        reqs_affected = os.path.join(dirname, "reqs_pinned_affected.txt")

        result = self.runner.invoke(self.cli, ['check', '-r', reqs_affected, '--db', db, '--output', 'html'])

        self.assertIn("remediations-suggested", result.stdout)
        self.assertIn("Use API Key", result.stdout)

    @patch('safety.safety.fetch_database_url')
    def test_license_with_file(self, fetch_database_url):
        licenses_db = {
            "licenses": {
                "BSD-3-Clause": 2
            },
            "packages": {
                "django": [
                    {
                        "start_version": "0.0",
                        "license_id": 2
                    }
                ]
            }
        }

        mock = Mock()
        mock.return_value = licenses_db

        dirname = os.path.dirname(__file__)
        test_filename = os.path.join(dirname, "reqs_4.txt")
        result = self.runner.invoke(self.cli, ['license', '--key', 'foo', '--file', test_filename])
        print(result.stdout)
        self.assertEqual(result.exit_code, 0)

    @patch('safety.auth.cli.get_auth_info', return_value={'email': 'test@test.com'})
    @patch.object(Auth, 'is_valid', return_value=True)
    @patch('safety.auth.utils.SafetyAuthSession.get_authentication_type', return_value=AuthenticationType.TOKEN)
    @patch('safety.safety.fetch_database', return_value={'vulnerable_packages': []})
    @patch('safety.auth.utils.initialize', return_value=None)
    @patch('safety.auth.cli_utils.SafetyCLI', return_value=SafetyCLI(platform_enabled=False, firewall_enabled=False))
    def test_debug_flag(self, *args):
        """
        Test the behavior of the CLI when invoked with the '--debug' flag.

        This test invokes the CLI with the 'scan' command and the '--debug' flag enabled,
        verifies that the command exits successfully, and checks that the expected output snippet
        is present in the CLI output.

        Args:
            mock_get_auth_info: Mock for retrieving authentication info.
            mock_is_valid: Mock for checking validity of inputs or authentication.
            mock_get_auth_type: Mock for retrieving the authentication type.
            mock_fetch_database: Mock for database fetching operations.
        """
        result = self.runner.invoke(self.cli, ['--debug', 'scan'])
        assert result.exit_code == 0, (
            f"CLI exited with code {result.exit_code} and output: {result.output} and error: {result.stderr}"
        )
        expected_output_snippet = f"{get_version()} scanning"
        assert expected_output_snippet in result.output, (
            f"Expected output to contain: {expected_output_snippet}, but got: {result.output}"
        )


    @patch('safety.auth.cli.get_auth_info', return_value={'email': 'test@test.com'})
    @patch.object(Auth, 'is_valid', return_value=True)
    @patch('safety.auth.utils.SafetyAuthSession.get_authentication_type', return_value=AuthenticationType.TOKEN)
    @patch('safety.safety.fetch_database', return_value={'vulnerable_packages': []})
    def test_debug_flag_with_value_1(self, *args):
        sys.argv = ['safety', '--debug', '1', 'scan']

        from safety.cli import preprocess_args

        @preprocess_args
        def dummy_function():
            pass

        # Extract the preprocessed arguments from sys.argv
        preprocessed_args = sys.argv[1:]  # Exclude the script name 'safety'

        # Assert the preprocessed arguments
        assert preprocessed_args == ['--debug', 'scan'], f"Preprocessed args: {preprocessed_args}"

    @patch('safety.auth.cli.get_auth_info', return_value={'email': 'test@test.com'})
    @patch.object(Auth, 'is_valid', return_value=True)
    @patch('safety.auth.utils.SafetyAuthSession.get_authentication_type', return_value=AuthenticationType.TOKEN)
    @patch('safety.safety.fetch_database', return_value={'vulnerable_packages': []})
    def test_debug_flag_with_value_true(self, *args):
        sys.argv = ['safety', '--debug', 'true', 'scan']

        from safety.cli import preprocess_args

        @preprocess_args
        def dummy_function():
            pass

        # Extract the preprocessed arguments from sys.argv
        preprocessed_args = sys.argv[1:]  # Exclude the script name 'safety'

        # Assert the preprocessed arguments
        assert preprocessed_args == ['--debug', 'scan'], f"Preprocessed args: {preprocessed_args}"

    @patch('safety.auth.utils.get_config_setting', return_value=None)
    @patch('safety.auth.cli.get_auth_info', return_value={'email': 'test@test.com'})
    @patch.object(Auth, 'is_valid', return_value=True)
    @patch('safety.auth.utils.SafetyAuthSession.get_authentication_type', return_value=AuthenticationType.TOKEN)
    @patch('safety.auth.utils.SafetyAuthSession.check_project', return_value={'user_confirm': True})
    @patch('safety.auth.utils.SafetyAuthSession.project', return_value={'slug': 'slug'})
    @patch('safety.auth.cli_utils.SafetyCLI', return_value=SafetyCLI(platform_enabled=True, firewall_enabled=True))
    def test_init_project(self, *args):
        # Workarounds
        from safety.console import main_console as test_console

        test_cases = [
            (False,
             "Configured PIP global settings\nConfigured PIP alias\n", 0),
            (True, 
             "Set a project id (no spaces). If empty Safety will use", 1)]

        for interactive, output, exit_code in test_cases:
            with self.subTest(output=output, exit_code=exit_code):
                with patch('safety.cli.console', 
                           new=test_console) as t_console, \
                     tempfile.TemporaryDirectory() as tempdir:
                    
                    t_console.is_interactive = interactive
                    
                    result = self.runner.invoke(self.cli, ['init', tempdir])
                    cleaned_stdout = click.unstyle(result.stdout)
                    assert result.exit_code == exit_code, f"CLI exited with non-zero exit code {result.exit_code}"
                    assert cleaned_stdout.startswith(output), f"CLI exited with output: {result.output}"


    @patch('safety.auth.cli.get_auth_info', return_value={'email': 'test@test.com'})
    @patch.object(Auth, 'is_valid', return_value=True)
    @patch('safety.auth.utils.SafetyAuthSession.get_authentication_type', return_value=AuthenticationType.TOKEN)
    @patch('safety.auth.utils.SafetyAuthSession.check_project', return_value={'user_confirm': True})
    @patch('safety.auth.utils.SafetyAuthSession.project', return_value={'slug': 'slug'})
    @patch('safety.auth.cli_utils.SafetyCLI', return_value=SafetyCLI(platform_enabled=True, firewall_enabled=True))
    def test_existing_project_is_linked(self, *args):
        dirname = os.path.dirname(__file__)
        source_project_file = os.path.join(dirname, "test-safety-project.ini")

        with tempfile.TemporaryDirectory() as tempdir:
            project_file = os.path.join(tempdir, '.safety-project.ini')
            shutil.copy(source_project_file, project_file)
            result = self.runner.invoke(self.cli, ['init', tempdir])
            assert result.exit_code == 0, f"CLI exited with code {result.exit_code} and output: {result.output} and error: {result.stderr}"

class TestNetworkTelemetry(unittest.TestCase):

    @patch('psutil.net_io_counters')
    @patch('psutil.net_if_addrs')
    @patch('psutil.net_connections')
    @patch('psutil.net_if_stats')
    @patch('requests.get')
    def test_get_network_telemetry(self, mock_requests_get, mock_net_if_stats, mock_net_connections, mock_net_if_addrs, mock_net_io_counters):
        # Setup mocks
        mock_net_io_counters.return_value = Mock(bytes_sent=1000, bytes_recv=2000, packets_sent=10, packets_recv=20)
        mock_net_if_addrs.return_value = {'eth0': [Mock(family=socket.AF_INET, address='192.168.1.1')]}
        mock_net_connections.return_value = [Mock(fd=1, family=socket.AF_INET, type=socket.SOCK_STREAM,
                                                  laddr=Mock(ip='192.168.1.1', port=8080),
                                                  raddr=Mock(ip='93.184.216.34', port=80),
                                                  status='ESTABLISHED')]
        mock_net_if_stats.return_value = {'eth0': Mock(isup=True, duplex=2, speed=1000, mtu=1500)}

        mock_response = Mock()
        mock_response.content = b'a' * 1000  # Mock content length
        mock_requests_get.return_value = mock_response

        # Run the function
        from safety.cli import get_network_telemetry
        result = get_network_telemetry()

        # Assert the network telemetry data
        self.assertEqual(result['bytes_sent'], 1000)
        self.assertEqual(result['bytes_recv'], 2000)
        self.assertEqual(result['packets_sent'], 10)
        self.assertEqual(result['packets_recv'], 20)
        self.assertIsNotNone(result['download_speed'])
        self.assertEqual(result['interfaces'], {'eth0': ['192.168.1.1']})
        self.assertEqual(result['connections'][0]['laddr'], '192.168.1.1:8080')
        self.assertEqual(result['connections'][0]['raddr'], '93.184.216.34:80')
        self.assertEqual(result['connections'][0]['status'], 'ESTABLISHED')
        self.assertEqual(result['interface_stats']['eth0']['isup'], True)
        self.assertEqual(result['interface_stats']['eth0']['duplex'], 2)
        self.assertEqual(result['interface_stats']['eth0']['speed'], 1000)
        self.assertEqual(result['interface_stats']['eth0']['mtu'], 1500)

    @patch('requests.get', side_effect=requests.RequestException('Network error'))
    def test_get_network_telemetry_request_exception(self, mock_requests_get):
        # Run the function
        from safety.cli import get_network_telemetry
        result = get_network_telemetry()

        # Assert the download_speed is None and error is captured
        self.assertIsNone(result['download_speed'])
        self.assertIn('error', result)

    @patch('psutil.net_io_counters', side_effect=psutil.AccessDenied('Access denied'))
    def test_get_network_telemetry_access_denied(self, mock_net_io_counters):
        # Run the function
        from safety.cli import get_network_telemetry
        result = get_network_telemetry()

        # Assert the error is captured
        self.assertIn('error', result)
        self.assertIn('Access denied', result['error'])

class TestConfigureLogger(unittest.TestCase):

    @patch('configparser.ConfigParser.read')
    @patch('safety.cli.get_network_telemetry')
    def test_configure_logger_debug(self, mock_get_network_telemetry, mock_config_read):
        mock_get_network_telemetry.return_value = {'dummy_key': 'dummy_value'}
        mock_config_read.return_value = None

        from safety.cli import configure_logger

        ctx = Mock()
        param = Mock()
        debug = True

        with patch('sys.argv', ['--debug', 'true']), \
        patch('logging.basicConfig') as mock_basicConfig, \
        patch('configparser.ConfigParser.items', return_value=[('key', 'value')]), \
        patch('configparser.ConfigParser.sections', return_value=['section']):
            configure_logger(ctx, param, debug)
            mock_basicConfig.assert_called_with(format='%(asctime)s %(name)s => %(message)s', level=logging.DEBUG)

        # Check if network telemetry logging was called
        mock_get_network_telemetry.assert_called_once()

    @patch('configparser.ConfigParser.read')
    def test_configure_logger_non_debug(self, mock_config_read):
        mock_config_read.return_value = None

        ctx = Mock()
        param = Mock()
        debug = False

        from safety.cli import configure_logger

        with patch('logging.basicConfig') as mock_basicConfig:
            configure_logger(ctx, param, debug)
        mock_basicConfig.assert_called_with(format='%(asctime)s %(name)s => %(message)s', level=logging.CRITICAL)



from pathlib import Path

from rich.prompt import Prompt
from ..cli_util import CommandType, FeatureType, SafetyCLICommand, SafetyCLISubGroup
from safety.error_handlers import handle_cmd_exception
import typer
import os


from safety.init.constants import (
    PROJECT_INIT_CMD_NAME,
    PROJECT_INIT_HELP,
    PROJECT_INIT_DIRECTORY_HELP,
)
from safety.init.main import create_project
from safety.console import main_console as console
from ..scan.command import scan
from ..scan.models import ScanOutput
from ..tool.main import (
    configure_system,
    configure_local_directory,
    has_local_tool_files,
    configure_alias,
)

from ..constants import CONTEXT_COMMAND_TYPE, CONTEXT_FEATURE_TYPE

from safety.decorators import notify
from safety.events.utils import emit_firewall_configured
from safety_schemas.models.events.payloads import AliasConfig, IndexConfig


try:
    from typing import Annotated  # type: ignore
except ImportError:
    from typing_extensions import Annotated

init_app = typer.Typer(rich_markup_mode="rich", cls=SafetyCLISubGroup)


@init_app.command(
    cls=SafetyCLICommand,
    help=PROJECT_INIT_HELP,
    name=PROJECT_INIT_CMD_NAME,
    options_metavar="[OPTIONS]",
    context_settings={
        "allow_extra_args": True,
        "ignore_unknown_options": True,
        CONTEXT_COMMAND_TYPE: CommandType.BETA,
        CONTEXT_FEATURE_TYPE: FeatureType.FIREWALL,
    },
)
@handle_cmd_exception
@notify
def init(
    ctx: typer.Context,
    directory: Annotated[
        Path,
        typer.Argument(  # type: ignore
            exists=True,
            file_okay=False,
            dir_okay=True,
            writable=False,
            readable=True,
            resolve_path=True,
            show_default=False,
            help=PROJECT_INIT_DIRECTORY_HELP,
        ),
    ] = Path("."),
):
    do_init(ctx, directory, False)


def do_init(ctx: typer.Context, directory: Path, prompt_user: bool = True):
    project_dir = (
        directory if os.path.isabs(directory) else os.path.join(os.getcwd(), directory)
    )

    # Fixed at the meantime, we'll revamp the init command soon.
    alias_configured = AliasConfig(is_configured=False)
    index_configured = IndexConfig(is_configured=False)

    create_project(ctx, console, Path(project_dir))

    answer = "y" if not prompt_user else None
    if prompt_user:
        console.print(
            "Safety prevents vulnerable or malicious packages from being installed on your computer. We do this by wrapping your package manager."
        )
        prompt = "Do you want to enable proactive malicious package prevention?"
        answer = Prompt.ask(
            prompt=prompt,
            choices=["y", "n"],
            default="y",
            show_default=True,
            console=console,
        ).lower()

    if answer == "y":
        configure_system(ctx.obj.org.get("slug"))
        # Naive approach, we'll revamp the init command soon.
        alias_configured.is_configured = True

    if prompt_user:
        prompt = "Do you want to alias pip to Safety?"
        answer = Prompt.ask(
            prompt=prompt,
            choices=["y", "n"],
            default="y",
            show_default=True,
            console=console,
        ).lower()

    if answer == "y":
        configure_alias()
        # Naive approach, we'll revamp the init command soon.
        index_configured.is_configured = True

    exit_code = None

    if has_local_tool_files(project_dir):  # type: ignore
        if prompt_user:
            prompt = "Do you want to enable proactive malicious package prevention for any project in working directory?"
            answer = Prompt.ask(
                prompt=prompt,
                choices=["y", "n"],
                default="y",
                show_default=True,
                console=console,
            ).lower()

        if answer == "y":
            configure_local_directory(
                project_dir,  # type: ignore
                ctx.obj.org.get("slug"),
                ctx.obj.project.id,
            )

        if prompt_user:
            prompt = "It looks like your current directory contains a requirements.txt file. Would you like Safety to scan it?"
            answer = Prompt.ask(
                prompt=prompt,
                choices=["y", "n"],
                default="y",
                show_default=True,
                console=console,
            ).lower()

        if answer == "y":
            ctx.command.name = "scan"
            ctx.params = {
                "target": directory,
                "output": ScanOutput.SCREEN,
                "policy_file_path": None,
            }
            try:
                scan(
                    ctx=ctx,
                    target=directory,
                    output=ScanOutput.SCREEN,
                    policy_file_path=None,
                )
            except SystemExit as e:
                exit_code = e

    emit_firewall_configured(
        event_bus=ctx.obj.event_bus,
        alias_config=alias_configured,
        index_config=index_configured,
    )

    if exit_code:
        raise exit_code



from typing import TYPE_CHECKING

from safety.events.handlers import EventHandler
from safety.events.types import EventBusReadyEvent
from safety.events.utils import emit_firewall_heartbeat

from safety.tool import ToolInspector

if TYPE_CHECKING:
    from safety.events.event_bus import EventBus


class HeartbeatInspectionEventHandler(EventHandler[EventBusReadyEvent]):
    """
    Inspect the system for installed tools and send an emit
    a firewall heartbeat event.
    """

    def __init__(self, event_bus: "EventBus") -> None:
        super().__init__()
        self.event_bus = event_bus

    async def handle(self, event: EventBusReadyEvent):
        ctx = event.payload.ctx
        inspector = ToolInspector(timeout=1.0)
        tools = await inspector.inspect_all_tools()

        emit_firewall_heartbeat(self.event_bus, ctx, tools=tools)



import logging
import uuid
import typer
from rich.console import Console

from .models import UnverifiedProjectModel

import configparser
from pathlib import Path
from safety_schemas.models import ProjectModel, Stage
from safety.scan.util import GIT
from ..auth.utils import SafetyAuthSession

from typing import Optional
from safety.scan.render import (
    print_wait_project_verification,
    prompt_project_id,
    prompt_link_project,
)

PROJECT_CONFIG = ".safety-project.ini"
PROJECT_CONFIG_SECTION = "project"
PROJECT_CONFIG_ID = "id"
PROJECT_CONFIG_URL = "url"
PROJECT_CONFIG_NAME = "name"


LOG = logging.getLogger(__name__)


def check_project(
    ctx: typer.Context,
    session: SafetyAuthSession,
    console: Console,
    unverified_project: UnverifiedProjectModel,
    git_origin: Optional[str],
    ask_project_id: bool = False,
) -> dict:
    """
    Check the project against the session and stage, verifying the project if necessary.

    Args:
        console: The console for output.
        ctx (typer.Context): The context of the Typer command.
        session (SafetyAuthSession): The authentication session.
        unverified_project (UnverifiedProjectModel): The unverified project model.
        stage (Stage): The current stage.
        git_origin (Optional[str]): The Git origin URL.
        ask_project_id (bool): Whether to prompt for the project ID.

    Returns:
        dict: The result of the project check.
    """
    stage = ctx.obj.auth.stage
    source = ctx.obj.telemetry.safety_source if ctx.obj.telemetry else None
    data = {"scan_stage": stage, "safety_source": source}

    PRJ_SLUG_KEY = "project_slug"
    PRJ_SLUG_SOURCE_KEY = "project_slug_source"
    PRJ_GIT_ORIGIN_KEY = "git_origin"

    if git_origin:
        data[PRJ_GIT_ORIGIN_KEY] = git_origin

    if unverified_project.id:
        data[PRJ_SLUG_KEY] = unverified_project.id
        data[PRJ_SLUG_SOURCE_KEY] = ".safety-project.ini"
    elif not git_origin or ask_project_id:
        default_id = unverified_project.project_path.parent.name

        if not default_id:
            # Sometimes the parent directory is empty, so we generate
            # a random ID
            default_id = str(uuid.uuid4())[:10]

        unverified_project.id = prompt_project_id(console, default_id)
        data[PRJ_SLUG_KEY] = unverified_project.id
        data[PRJ_SLUG_SOURCE_KEY] = "user"

    status = print_wait_project_verification(
        console,
        data[PRJ_SLUG_KEY] if data.get(PRJ_SLUG_KEY, None) else "-",
        (session.check_project, data),
        on_error_delay=1,
    )

    return status


def verify_project(
    console: Console,
    ctx: typer.Context,
    session: SafetyAuthSession,
    unverified_project: UnverifiedProjectModel,
    stage: Stage,
    git_origin: Optional[str],
):
    """
    Verify the project, linking it if necessary and saving the verified project information.

    Args:
        console: The console for output.
        ctx (typer.Context): The context of the Typer command.
        session (SafetyAuthSession): The authentication session.
        unverified_project (UnverifiedProjectModel): The unverified project model.
        stage (Stage): The current stage.
        git_origin (Optional[str]): The Git origin URL.
    """

    verified_prj = False

    link_prj = True

    while not verified_prj:
        result = check_project(
            ctx,
            session,
            console,
            unverified_project,
            git_origin,
            ask_project_id=not link_prj,
        )

        unverified_slug = result.get("slug")

        project = result.get("project", None)
        user_confirm = result.get("user_confirm", False)

        if user_confirm:
            if project and link_prj:
                prj_name = project.get("name", None)
                prj_admin_email = project.get("admin", None)

                link_prj = prompt_link_project(
                    prj_name=prj_name, prj_admin_email=prj_admin_email, console=console
                )

                if not link_prj:
                    continue

        verified_prj = print_wait_project_verification(
            console,
            unverified_slug,  # type: ignore
            (session.project, {"project_id": unverified_slug}),
            on_error_delay=1,
        )

        if (
            verified_prj
            and isinstance(verified_prj, dict)
            and verified_prj.get("slug", None)
        ):
            save_verified_project(
                ctx,
                verified_prj["slug"],
                verified_prj.get("name", None),
                unverified_project.project_path,
                verified_prj.get("url", None),
                verified_prj.get("organization", None),
            )
        else:
            verified_prj = False


def load_unverified_project_from_config(project_root: Path) -> UnverifiedProjectModel:
    """
    Loads an unverified project from the configuration file located at the project root.

    Args:
        project_root (Path): The root directory of the project.

    Returns:
        UnverifiedProjectModel: An instance of UnverifiedProjectModel.
    """
    config = configparser.ConfigParser()
    project_path = project_root / PROJECT_CONFIG
    config.read(project_path)
    id = config.get(PROJECT_CONFIG_SECTION, PROJECT_CONFIG_ID, fallback=None)
    url = config.get(PROJECT_CONFIG_SECTION, PROJECT_CONFIG_URL, fallback=None)
    name = config.get(PROJECT_CONFIG_SECTION, PROJECT_CONFIG_NAME, fallback=None)
    created = True
    if id:
        created = False

    return UnverifiedProjectModel(
        id=id, url_path=url, name=name, project_path=project_path, created=created
    )


def save_verified_project(
    ctx: typer.Context,
    slug: str,
    name: Optional[str],
    project_path: Path,
    url_path: Optional[str],
    organization: Optional[dict],
):
    """
    Save the verified project information to the context and project info file.

    Args:
        ctx (typer.Context): The context of the Typer command.
        slug (str): The project slug.
        name (Optional[str]): The project name.
        project_path (Path): The project path.
        url_path (Optional[str]): The project URL path.
        organization (Optional[str]): The project organization.
    """
    ctx.obj.project = ProjectModel(
        id=slug, name=name, project_path=project_path, url_path=url_path
    )

    save_project_info(project=ctx.obj.project, project_path=project_path)

    ctx.obj.org = {}
    if organization:
        ctx.obj.org = {
            "name": organization.get("name"),
            "slug": organization.get("slug"),
        }


def save_project_info(project: ProjectModel, project_path: Path) -> bool:
    """
    Saves the project information to the configuration file.

    Args:
        project (ProjectModel): The ProjectModel object containing project
                                information.
        project_path (Path): The path to the configuration file.

    Returns:
        bool: True if the project information was saved successfully, False
              otherwise.
    """
    config = configparser.ConfigParser()
    config.read(project_path)

    if PROJECT_CONFIG_SECTION not in config.sections():
        config[PROJECT_CONFIG_SECTION] = {}

    config[PROJECT_CONFIG_SECTION][PROJECT_CONFIG_ID] = project.id
    if project.url_path:
        config[PROJECT_CONFIG_SECTION][PROJECT_CONFIG_URL] = project.url_path
    if project.name:
        config[PROJECT_CONFIG_SECTION][PROJECT_CONFIG_NAME] = project.name

    try:
        with open(project_path, "w") as configfile:
            config.write(configfile)
    except Exception:
        LOG.exception("Error saving project info")
        return False

    return True


def create_project(ctx: typer.Context, console: Console, target: Path):
    """
    Loads existing project from the specified target locations or creates a new project.

    Args:
        ctx: The CLI context
        session: The authentication session
        console: The console object
        target (Path): The target location
    """
    # Load .safety-project.ini
    unverified_project = load_unverified_project_from_config(project_root=target)

    stage = ctx.obj.auth.stage
    session = ctx.obj.auth.client
    git_data = GIT(root=target).build_git_data()
    origin = None

    if git_data:
        origin = git_data.origin

    if ctx.obj.platform_enabled:
        verify_project(console, ctx, session, unverified_project, stage, origin)
    else:
        console.print("Project creation is not supported for your account.")



from importlib.metadata import PackageNotFoundError, version
import logging
from typing import Optional


LOG = logging.getLogger(__name__)


def get_version() -> Optional[str]:
  """
  Get the version of the Safety package.

  Returns:
    Optional[str]: The Safety version if found, otherwise None.
  """
  try:
    return version("safety")
  except PackageNotFoundError:
    LOG.exception("Unable to get Safety version.")
    return None



from sys import platform
from typing import Optional
from .types import InterceptorType
from .unix import UnixAliasInterceptor
from .windows import WindowsInterceptor
from .base import CommandInterceptor


def create_interceptor(
    interceptor_type: Optional[InterceptorType] = None,
) -> CommandInterceptor:
    """
    Create appropriate interceptor based on OS and type
    """
    interceptor_map = {
        InterceptorType.UNIX_ALIAS: UnixAliasInterceptor,
        InterceptorType.WINDOWS_BAT: WindowsInterceptor,
    }

    if interceptor_type:
        return interceptor_map[interceptor_type]()

    # Auto-select based on OS
    if platform == "win32":
        return interceptor_map[InterceptorType.WINDOWS_BAT]()

    if platform in ["linux", "linux2", "darwin"]:
        # Default to alias-based on Unix-like systems
        return interceptor_map[InterceptorType.UNIX_ALIAS]()

    raise NotImplementedError(f"Platform '{platform}' is not supported.")



from functools import wraps
from pathlib import Path

from safety_schemas.models import ProjectModel

from safety.console import main_console

from ..cli_util import process_auth_status_not_ready
from ..init.main import load_unverified_project_from_config, verify_project
from ..scan.util import GIT


def optional_project_command(func):
    @wraps(func)
    def inner(ctx, target: Path, *args, **kwargs):
        ctx.obj.console = main_console
        ctx.params.pop("console", None)

        if not ctx.obj.auth.is_valid():
            process_auth_status_not_ready(
                console=main_console, auth=ctx.obj.auth, ctx=ctx
            )

        upload_request_id = kwargs.pop("upload_request_id", None)

        # Load .safety-project.ini
        unverified_project = load_unverified_project_from_config(project_root=target)

        if ctx.obj.platform_enabled and not unverified_project.created:
            stage = ctx.obj.auth.stage
            session = ctx.obj.auth.client
            git_data = GIT(root=target).build_git_data()
            origin = None

            if git_data:
                origin = git_data.origin

            verify_project(
                main_console, ctx, session, unverified_project, stage, origin
            )

            ctx.obj.project.git = git_data
        else:
            ctx.obj.project = ProjectModel(
                id="",
                name="Undefined project",
                project_path=unverified_project.project_path,
            )

        ctx.obj.project.upload_request_id = upload_request_id

        return func(ctx, target=target, *args, **kwargs)

    return inner


def register_event_handlers(func):
    @wraps(func)
    def inner(ctx, *args, **kwargs):
        if event_bus := ctx.obj.event_bus:
            from safety_schemas.models.events import EventType

            event_bus.subscribe(
                [
                    EventType.PACKAGE_UNINSTALLED,
                    EventType.PACKAGE_INSTALLED,
                    EventType.PACKAGE_UPDATED,
                    EventType.TOOL_COMMAND_EXECUTED,
                ],
                ctx.obj.security_events_handler,
            )

        return func(ctx, *args, **kwargs)

    return inner



import logging
from typing import Dict, Optional, Tuple, Any, Callable

import click

from .main import get_auth_info, get_host_config, get_organization, get_proxy_config, \
    get_redirect_url, get_token_data, save_auth_config, get_token, clean_session
from authlib.common.security import generate_token
from safety.auth.constants import CLIENT_ID, OPENID_CONFIG_URL

from safety.auth.models import Organization, Auth
from safety.auth.utils import S3PresignedAdapter, SafetyAuthSession, get_keys, is_email_verified
from safety.constants import REQUEST_TIMEOUT
from safety.scan.constants import CLI_KEY_HELP, CLI_PROXY_HOST_HELP, CLI_PROXY_PORT_HELP, CLI_PROXY_PROTOCOL_HELP, CLI_STAGE_HELP
from safety.util import DependentOption, SafetyContext, get_proxy_dict
from safety.models import SafetyCLI
from safety_schemas.models import Stage

LOG = logging.getLogger(__name__)


def build_client_session(api_key: Optional[str] = None, proxies: Optional[Dict[str, str]] = None, headers: Optional[Dict[str, str]] = None) -> Tuple[SafetyAuthSession, Dict[str, Any]]:
    """
    Builds and configures the client session for authentication.

    Args:
        api_key (Optional[str]): The API key for authentication.
        proxies (Optional[Dict[str, str]]): Proxy configuration.
        headers (Optional[Dict[str, str]]): Additional headers.

    Returns:
        Tuple[SafetyAuthSession, Dict[str, Any]]: The configured client session and OpenID configuration.
    """

    kwargs = {}
    target_proxies = proxies

    # Global proxy defined in the config.ini
    proxy_config, proxy_timeout, proxy_required = get_proxy_config()

    if not proxies:
        target_proxies = proxy_config

    def update_token(tokens, **kwargs):
        save_auth_config(access_token=tokens['access_token'], id_token=tokens['id_token'],
                    refresh_token=tokens['refresh_token'])
        load_auth_session(click_ctx=click.get_current_context(silent=True))

    client_session = SafetyAuthSession(client_id=CLIENT_ID,
                                       code_challenge_method='S256',
                                       redirect_uri=get_redirect_url(),
                                       update_token=update_token,
                                       scope='openid email profile offline_access',
                                       **kwargs)

    client_session.mount("https://pyup.io/static-s3/", S3PresignedAdapter())

    client_session.proxy_required = proxy_required
    client_session.proxy_timeout = proxy_timeout
    client_session.proxies = target_proxies
    client_session.headers = {"Accept": "application/json", "Content-Type": "application/json"}

    try:
        openid_config = client_session.get(url=OPENID_CONFIG_URL, timeout=REQUEST_TIMEOUT).json()
    except Exception as e:
        LOG.debug('Unable to load the openID config: %s', e)
        openid_config = {}

    client_session.metadata["token_endpoint"] = openid_config.get("token_endpoint",
                                                                  None)

    if api_key:
        client_session.api_key = api_key
        client_session.headers['X-Api-Key'] = api_key

    if headers:
        client_session.headers.update(headers)

    return client_session, openid_config


def load_auth_session(click_ctx: click.Context) -> None:
    """
    Loads the authentication session from the context.

    Args:
        click_ctx (click.Context): The Click context object.
    """
    if not click_ctx:
        LOG.warning("Click context is needed to be able to load the Auth data.")
        return

    client = click_ctx.obj.auth.client
    keys = click_ctx.obj.auth.keys

    access_token: str = get_token(name='access_token')
    refresh_token: str = get_token(name='refresh_token')
    id_token: str = get_token(name='id_token')

    if access_token and keys:
        try:
            token = get_token_data(access_token, keys, silent_if_expired=True)
            client.token = {'access_token': access_token,
                                    'refresh_token': refresh_token,
                                    'id_token': id_token,
                                    'token_type': 'bearer',
                                    'expires_at': token.get('exp', None)}
        except Exception as e:
            print(e)
            clean_session(client)

def proxy_options(func: Callable) -> Callable:
    """
    Decorator that defines proxy options for Click commands.

    Options defined per command, this will override the proxy settings defined in the
    config.ini file.

    Args:
        func (Callable): The Click command function.

    Returns:
        Callable: The wrapped Click command function with proxy options.
    """
    func = click.option("--proxy-protocol",
                        type=click.Choice(['http', 'https']), default='https',
                        cls=DependentOption, required_options=['proxy_host'],
                        help=CLI_PROXY_PROTOCOL_HELP)(func)
    func = click.option("--proxy-port", multiple=False, type=int, default=80,
                        cls=DependentOption, required_options=['proxy_host'],
                        help=CLI_PROXY_PORT_HELP)(func)
    func = click.option("--proxy-host", multiple=False, type=str, default=None,
                        help=CLI_PROXY_HOST_HELP)(func)

    return func

def auth_options(stage: bool = True) -> Callable:
    """
    Decorator that defines authentication options for Click commands.

    Args:
        stage (bool): Whether to include the stage option.

    Returns:
        Callable: The decorator function.
    """
    def decorator(func: Callable) -> Callable:

        func = click.option("--key", default=None, envvar="SAFETY_API_KEY",
            help=CLI_KEY_HELP)(func)

        if stage:
            func = click.option("--stage", default=None, envvar="SAFETY_STAGE",
                                help=CLI_STAGE_HELP)(func)

        return func

    return decorator


def inject_session(ctx: click.Context, proxy_protocol: Optional[str] = None,
            proxy_host: Optional[str] = None,
            proxy_port: Optional[str] = None,
            key: Optional[str] = None,
            stage: Optional[Stage] = None,
            invoked_command: str = "") -> Any:

    org: Optional[Organization] = get_organization()

    if not stage:
        host_stage = get_host_config(key_name="stage")
        stage = host_stage if host_stage else Stage.development

    proxy_config: Optional[Dict[str, str]] = get_proxy_dict(proxy_protocol,
                                                            proxy_host, proxy_port)

    client_session, openid_config = build_client_session(api_key=key,
                                                            proxies=proxy_config)
    keys = get_keys(client_session, openid_config)

    auth = Auth(
        stage=stage,
        keys=keys,
        org=org,
        client_id=CLIENT_ID,
        client=client_session,
        code_verifier=generate_token(48)
    )

    if not ctx.obj:        
        ctx.obj = SafetyCLI()

    ctx.obj.auth = auth

    load_auth_session(ctx)

    info = get_auth_info(ctx)

    if info:
        ctx.obj.auth.name = info.get("name")
        ctx.obj.auth.email = info.get("email")
        ctx.obj.auth.email_verified = is_email_verified(info)
        SafetyContext().account = info["email"]
    else:
        SafetyContext().account = ""

    @ctx.call_on_close
    def clean_up_on_close():
        LOG.debug('Closing requests session.')
        ctx.obj.auth.client.close()

        if ctx.obj.event_bus:
            from safety.events.utils import create_internal_event, \
                InternalEventType, InternalPayload

            payload = InternalPayload(ctx=ctx)

            flush_event = create_internal_event(
                event_type=InternalEventType.FLUSH_SECURITY_TRACES,
                payload=payload
            )
            close_event = create_internal_event(
                event_type=InternalEventType.CLOSE_RESOURCES,
                payload=payload
            )
            
            flush_future = ctx.obj.event_bus.emit(flush_event)
            close_future = ctx.obj.event_bus.emit(close_event)
            
            # Wait for both events to be processed
            if flush_future and close_future:
                try:
                    flush_future.result()
                    close_future.result()
                except Exception as e:
                    LOG.warning(f"Error waiting for events to process: {e}")

            ctx.obj.event_bus.stop()



from datetime import datetime
from pathlib import Path
from typing import Dict, List, Literal, Optional, Union

from pydantic import Field, HttpUrl

from common.schemas import BaseModel, ConstrainedDict
from scans.schemas.base import (
    GitInfo,
    NoGit,
    PackageShort,
    RemediationsResults,
    RequirementInfo,
    Telemetry,
    Vulnerability,
)


class Meta(BaseModel):
    """
    Metadata for the scan report.

    Attributes:
        scan_type (Literal["system-scan", "scan", "check"]): The type of scan.
        scan_location (Path): The location of the scan.
        logged_to_dashboard (bool): Whether the scan was logged to the dashboard.
        authenticated (bool): Whether the scan was authenticated.
        authentication_method (Literal["token", "api_key"]): The method of authentication.
        local_database_path (Optional[Path]): The path to the local database.
        safety_version (str): The version of the Safety tool used.
        timestamp (datetime): The timestamp of the scan.
        telemetry (Telemetry): Telemetry data related to the scan.
        schema_version (str): The version of the schema used.
    """
    scan_type: Literal["system-scan", "scan", "check"]
    scan_location: Path
    logged_to_dashboard: bool
    authenticated: bool
    authentication_method: Literal["token", "api_key"]
    local_database_path: Optional[Path]
    safety_version: str
    timestamp: datetime
    telemetry: Telemetry
    schema_version: str


class Package(BaseModel):
    """
    Information about a package and its vulnerabilities.

    Attributes:
        requirements (ConstrainedDict[str, RequirementInfo]): The package requirements.
        current_version (Optional[str]): The current version of the package.
        vulnerabilities_found (Optional[int]): The number of vulnerabilities found.
        recommended_version (Optional[str]): The recommended version of the package.
        other_recommended_versions (List[str]): Other recommended versions of the package.
        more_info_url (Optional[HttpUrl]): URL for more information about the package.
    """
    requirements: ConstrainedDict[str, RequirementInfo]
    current_version: Optional[str]
    vulnerabilities_found: Optional[int]
    recommended_version: Optional[str]
    other_recommended_versions: List[str] = Field([], max_items=100, unique_items=True)
    more_info_url: Optional[HttpUrl]


class OSVulnerabilities(BaseModel):
    """
    Information about OS vulnerabilities.

    Attributes:
        packages (ConstrainedDict[str, Package]): Packages with vulnerabilities.
        vulnerabilities (List[Vulnerability]): List of vulnerabilities.
    """
    packages: ConstrainedDict[str, Package]
    vulnerabilities: List[Vulnerability] = Field(..., max_items=100, unique_items=True)


class EnvironmentFindings(BaseModel):
    """
    Findings related to the environment.

    Attributes:
        configuration (ConstrainedDict): Configuration details.
        packages (ConstrainedDict[str, Package]): Packages found in the environment.
        os_vulnerabilities (OSVulnerabilities): OS vulnerabilities found.
    """
    configuration: ConstrainedDict
    packages: ConstrainedDict[str, Package]
    os_vulnerabilities: OSVulnerabilities


class Environment(BaseModel):
    """
    Details about the environment being scanned.

    Attributes:
        full_location (Path): The full path of the environment.
        type (Literal["environment"]): The type of the environment.
        findings (EnvironmentFindings): Findings related to the environment.
    """
    full_location: Path
    type: Literal["environment"]
    findings: EnvironmentFindings


class DependencyVulnerabilities(BaseModel):
    """
    Information about dependency vulnerabilities.

    Attributes:
        packages (List[PackageShort]): List of packages with vulnerabilities.
        vulnerabilities (List[Vulnerability]): List of vulnerabilities found.
    """
    packages: List[PackageShort] = Field(..., max_items=500, unique_items=True)
    vulnerabilities: List[Vulnerability] = Field(..., max_items=100, unique_items=True)


class FileFindings(BaseModel):
    """
    Findings related to a file.

    Attributes:
        configuration (ConstrainedDict): Configuration details.
        packages (List[PackageShort]): List of packages found in the file.
        dependency_vulnerabilities (DependencyVulnerabilities): Dependency vulnerabilities found.
    """
    configuration: ConstrainedDict
    packages: List[PackageShort] = Field(..., max_items=500, unique_items=True)
    dependency_vulnerabilities: DependencyVulnerabilities


class Remediations(BaseModel):
    """
    Remediations for vulnerabilities.

    Attributes:
        configuration (ConstrainedDict): Configuration details.
        packages (ConstrainedDict[str, Package]): Packages with remediations.
        dependency_vulnerabilities (ConstrainedDict[str, Package]): Dependency vulnerabilities with remediations.
        remediations_results (RemediationsResults): Results of the remediations.
    """
    configuration: ConstrainedDict
    packages: ConstrainedDict[str, Package]
    dependency_vulnerabilities: ConstrainedDict[str, Package]
    remediations_results: RemediationsResults


class File(BaseModel):
    """
    Information about a scanned file.

    Attributes:
        full_location (Path): The full path of the file.
        type (str): The type of the file.
        language (Literal["python"]): The programming language of the file.
        format (str): The format of the file.
        findings (FileFindings): Findings related to the file.
        remediations (Remediations): Remediations for the file.
    """
    full_location: Path
    type: str
    language: Literal["python"]
    format: str
    findings: FileFindings
    remediations: Remediations


class Results(BaseModel):
    """
    The results of a scan.

    Attributes:
        environments (List[ConstrainedDict[Path, Environment]]): List of environments scanned.
        files (List[ConstrainedDict[str, File]]): List of files scanned.
    """
    environments: List[ConstrainedDict[Path, Environment]] = Field(
        [], max_items=100, unique_items=True
    )
    files: List[ConstrainedDict[str, File]] = Field(
        [], max_items=100, unique_items=True
    )


class Project(Results):
    """
    Information about a project being scanned.

    Attributes:
        id (Optional[int]): The project ID.
        location (Path): The location of the project.
        policy (Optional[Path]): The policy file for the project.
        policy_source (Optional[Literal["local", "cloud"]]): The source of the policy.
        git (Union[GitInfo, NoGit]): Git information related to the project.
    """
    id: Optional[int]
    location: Path
    policy: Optional[Path]
    policy_source: Optional[Literal["local", "cloud"]]
    git: Union[GitInfo, NoGit]


class ScanReportV30(BaseModel):
    """
    The scan report.

    Attributes:
        meta (Meta): Metadata about the scan.
        results (Union[Results, Dict]): The results of the scan.
        projects (Union[Project, Dict]): Projects involved in the scan.
    """
    meta: Meta
    results: Results | Dict = {}
    projects: Project | Dict = {}


import logging
import os
from rich.console import Console
from rich.theme import Theme

LOG = logging.getLogger(__name__)

SAFETY_THEME = {
    "file_title": "bold default on default",
    "dep_name": "bold yellow on default",
    "scan_meta_title": "bold default on default",
    "vuln_brief": "red on default",
    "rem_brief": "bold green on default",
    "rem_severity": "bold red on default",
    "brief_severity": "bold default on default",
    "status.spinner": "green",
    "recommended_ver": "bold cyan on default",
    "vuln_id": "bold default on default",
    "number": "bold cyan on default",
    "link": "underline bright_blue on default",
    "tip": "bold default on default",
    "specifier": "bold cyan on default",
    "vulns_found_number": "red on default",
}

non_interactive = os.getenv('NON_INTERACTIVE') == '1'

console_kwargs = {"theme": Theme(SAFETY_THEME, inherit=False)}

if non_interactive:
    LOG.info("NON_INTERACTIVE environment variable is set, forcing non-interactive mode")
    console_kwargs.update({"force_terminal": True, "force_interactive": False})


main_console = Console(**console_kwargs)
